# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the debusine Cli Collection commands."""

import re
from functools import partial
from typing import override
from unittest import mock

import yaml

from debusine.artifacts.models import CollectionCategory
from debusine.client.commands.base import Command, DebusineCommand
from debusine.client.commands.collections import (
    CollectionCommand,
    SingleCollectionCommand,
)
from debusine.client.commands.tests.base import BaseCliTests
from debusine.client.exceptions import DebusineError
from debusine.client.models import (
    CollectionData,
    CollectionDataNew,
    CollectionItemResponse,
    LookupChildType,
    LookupMultipleResponse,
    LookupResultType,
    LookupSingleResponse,
)


class CollectionCommandTests(BaseCliTests):
    """Tests for the :py:class:`CollectionCommand` class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.data = CollectionData(
            id=42,
            name="sid",
            category=CollectionCategory.SUITE,
            full_history_retention_period=7,
            metadata_only_retention_period=14,
            data={"test": 2},
            links={
                "webui_self": "/c/42",
                "webui_category": "/c/suite",
            },
        )

    @Command.preserve_registry()
    def _command(self) -> CollectionCommand:
        class ConcreteCollectionCommand(CollectionCommand):
            """Version of CollectionCommand that can be instantiated."""

            @override
            def run(self) -> None:
                raise NotImplementedError()

        return self.command(ConcreteCollectionCommand, workspace="workspace")

    def test_markup_category(self) -> None:
        command = self._command()
        self.assertEqual(
            command.markup_category(self.data),
            "[link=https://debusine.debian.org/c/suite]debian:suite[/]",
        )
        self.data.category = CollectionCategory.WORKFLOW_INTERNAL
        self.assertEqual(
            command.markup_category(self.data), "debusine:workflow-internal"
        )

    def test_list_rich(self) -> None:
        command = self._command()
        with (
            mock.patch(
                "debusine.client.commands.collections.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.collections.rich.print"
            ) as rich_print,
        ):
            stderr, stdout = self.capture_output(
                partial(command._list_rich, [self.data])
            )

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        rich_print.assert_called_once()

        self.assertEqual(table.add_row.call_count, 1)
        url = "https://debusine.debian.org/c/42"
        category_url = "https://debusine.debian.org/c/suite"
        call = table.add_row.call_args_list[0]
        self.assertEqual(call.args[0], "42")
        self.assertEqual(call.args[1], f"[link={url}]sid[/]")
        self.assertEqual(call.args[2], f"[link={category_url}]debian:suite[/]")
        self.assertEqual(call.args[3], "7")
        self.assertEqual(call.args[4], "14")

    def test_show_rich(self) -> None:
        command = self._command()
        with (
            mock.patch(
                "debusine.client.commands.collections.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.collections.rich.print"
            ) as rich_print,
            mock.patch("rich.json.JSON.from_data", side_effect=lambda x: x),
        ):
            stderr, stdout = self.capture_output(
                partial(command._show_rich, self.data)
            )

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        rich_print.assert_called_once()

        self.assertEqual(table.add_row.call_count, 7)
        url = "https://debusine.debian.org/c/42"
        category_url = "https://debusine.debian.org/c/suite"
        rows = [
            call_args_list.args
            for call_args_list in table.add_row.call_args_list
        ]
        self.assertEqual(rows[0], ("ID:", "#42"))
        self.assertEqual(rows[1], ("Web URL:", f"[link={url}]{url}[/]"))
        self.assertEqual(rows[2], ("Name:", "sid"))
        self.assertEqual(
            rows[3], ("Category:", f"[link={category_url}]debian:suite[/]")
        )
        self.assertEqual(rows[4], ("Full history retention:", "7 days"))
        self.assertEqual(rows[5], ("Metadata only retention:", "14 days"))
        self.assertEqual(rows[6], ("Data:", self.data.data))

    def test_show_rich_stats(self) -> None:
        self.data.stats = [
            {"type": "test", "category": "test", "active": True, "count": 2}
        ]
        command = self._command()
        with (
            mock.patch(
                "debusine.client.commands.collections.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.collections.rich.print"
            ) as rich_print,
            mock.patch("rich.json.JSON.from_data", side_effect=lambda x: x),
        ):
            stderr, stdout = self.capture_output(
                partial(command._show_rich, self.data)
            )

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        rich_print.assert_called_once()

        self.assertEqual(table.add_row.call_count, 8)
        url = "https://debusine.debian.org/c/42"
        category_url = "https://debusine.debian.org/c/suite"
        rows = [
            call_args_list.args
            for call_args_list in table.add_row.call_args_list
        ]
        self.assertEqual(rows[0], ("ID:", "#42"))
        self.assertEqual(rows[1], ("Web URL:", f"[link={url}]{url}[/]"))
        self.assertEqual(rows[2], ("Name:", "sid"))
        self.assertEqual(
            rows[3], ("Category:", f"[link={category_url}]debian:suite[/]")
        )
        self.assertEqual(rows[4], ("Full history retention:", "7 days"))
        self.assertEqual(rows[5], ("Metadata only retention:", "14 days"))
        self.assertEqual(rows[6], ("Data:", self.data.data))
        self.assertEqual(rows[7], ("Stats:", self.data.stats))


class SingleCollectionCommandTests(BaseCliTests):
    """Tests for the :py:class:`SingleCollectionCommand` class."""

    @Command.preserve_registry()
    def _command(
        self, category_or_fullname: str, name: str | None = None
    ) -> SingleCollectionCommand:
        class ConcreteCollectionCommand(SingleCollectionCommand):
            """Version of SingleCollectionCommand that can be instantiated."""

            @override
            def run(self) -> None:
                raise NotImplementedError()

        return self.command(
            ConcreteCollectionCommand,
            workspace="workspace",
            category_or_fullname=category_or_fullname,
            name=name,
        )

    def test_missing_name(self) -> None:
        for arg in ("debian:suite", "debian:archive", "invalid"):
            with self.subTest(arg=arg):
                command = self._command(arg)
                self.assertEqual(command.collection_category, arg)
                self.assertEqual(command.collection_name, "_")

    def test_full_name(self) -> None:
        for arg in (
            "sid@debian:suite",
            "_@debian:suite",
            "sid@debian:archive",
            "_@debian:archive",
        ):
            with self.subTest(arg=arg):
                command = self._command(arg)
                name, category = arg.split("@")
                self.assertEqual(command.collection_category, category)
                self.assertEqual(command.collection_name, name)

    def test_separate_name(self) -> None:
        for category, name in (
            ("debian:archive", "_"),
            ("debian:suite", "_"),
            ("debian:archive", "sid"),
            ("debian:suite", "sid"),
        ):
            with self.subTest(category=category, name=name):
                command = self._command(category, name)
                self.assertEqual(command.collection_category, category)
                self.assertEqual(command.collection_name, name)

    def test_name_given_twice(self) -> None:
        for category, name in (
            ("_@debian:archive", "_"),
            ("_@debian:suite", "_"),
            ("sid@debian:archive", "sid"),
            ("sid@debian:suite", "sid"),
            ("a@debian:suite", "b"),
            ("a@debian:suite", "b"),
        ):
            with self.subTest(category=category, name=name):
                stderr, _ = self.capture_output(
                    partial(self._command, category, name),
                    assert_system_exit_code=2,
                )
                self.assertRegex(
                    stderr,
                    "collection name is given"
                    f" both in `{re.escape(category)}`"
                    " and in the name argument",
                )


class ListTests(BaseCliTests):
    """Tests for the :py:class:`collections.List` class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.data = CollectionData(
            id=42,
            name="sid",
            category=CollectionCategory.SUITE,
            full_history_retention_period=7,
            metadata_only_retention_period=14,
            data={"test": 2},
        )

    def test_list(self) -> None:
        command = self.create_command(
            ["collection", "list", "--workspace=workspace"]
        )
        assert isinstance(command, DebusineCommand)
        with (
            mock.patch.object(
                command.debusine, "collection_iter", return_value=[self.data]
            ) as collection_iter,
            mock.patch.object(command, "list") as list_,
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        collection_iter.assert_called_once_with("workspace")
        list_.assert_called_once_with([self.data])


class ShowTests(BaseCliTests):
    """Tests for the :py:class:`collections.Show` class."""

    def test_show(self) -> None:
        data = CollectionData(
            id=42,
            name="sid",
            category=CollectionCategory.SUITE,
            full_history_retention_period=7,
            metadata_only_retention_period=14,
            data={"test": 2},
        )

        command = self.create_command(
            ["collection", "show", "--workspace=workspace", "category", "name"]
        )
        assert isinstance(command, DebusineCommand)
        with (
            mock.patch.object(
                command.debusine, "collection_get", return_value=data
            ) as collection_get,
            mock.patch.object(command, "show") as show,
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        collection_get.assert_called_once_with(
            "workspace", "category", "name", with_stats=True
        )
        show.assert_called_once_with(data)


class CreateTests(BaseCliTests):
    """Tests for the :py:class:`collections.Create` class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.sent = CollectionDataNew(
            name="sid",
            category=CollectionCategory.SUITE,
            data={},
        )

    def assertCreates(self, *args: str) -> None:
        """Collection_create with the given arguments returns self.sample."""
        command = self.create_command(
            [
                "collection",
                "create",
                "--workspace=workspace",
                CollectionCategory.SUITE,
                "sid",
                *args,
            ]
        )
        assert isinstance(command, DebusineCommand)

        created = CollectionData(id=42, **self.sent.model_dump())
        with (
            mock.patch.object(
                command.debusine, "collection_create", return_value=created
            ) as collection_create,
            mock.patch.object(command, "show") as show,
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        collection_create.assert_called_once_with("workspace", self.sent)
        show.assert_called_once_with(created)

    def test_defaults(self) -> None:
        self.enterContext(self.patch_sys_stdin_read(""))
        self.assertCreates()

    def test_set_full_history_retention_period(self) -> None:
        self.enterContext(self.patch_sys_stdin_read(""))
        self.sent.full_history_retention_period = 1
        self.assertCreates("--full-history-retention-period=1")

    def test_unset_full_history_retention_period(self) -> None:
        self.enterContext(self.patch_sys_stdin_read(""))
        self.sent.full_history_retention_period = None
        self.assertCreates("--full-history-retention-period=-")

    def test_set_metadata_only_retention_period(self) -> None:
        self.enterContext(self.patch_sys_stdin_read(""))
        self.sent.metadata_only_retention_period = 1
        self.assertCreates("--metadata-only-retention-period=1")

    def test_unset_metadata_only_retention_period(self) -> None:
        self.enterContext(self.patch_sys_stdin_read(""))
        self.sent.metadata_only_retention_period = None
        self.assertCreates("--metadata-only-retention-period=-")

    def test_set_data(self) -> None:
        infile = self.create_temporary_file()
        infile.write_text("{'test': 7}")
        self.sent.data = {"test": 7}
        self.assertCreates(f"--data={infile}")

    def test_set_data_stdin(self) -> None:
        self.enterContext(self.patch_sys_stdin_read("{'test': 7}"))
        self.sent.data = {"test": 7}
        self.assertCreates("--data=-")

    def test_invalid_period(self) -> None:
        stderr, _ = self.capture_output(
            partial(
                self.create_command,
                [
                    "collection",
                    "create",
                    "--workspace=workspace",
                    "category",
                    "name",
                    "--full-history-retention-period=invalid",
                ],
            ),
            assert_system_exit_code=2,
        )
        self.assertRegex(
            stderr,
            r"error: retention period must be a number or '-'",
        )


class ManageTests(BaseCliTests):
    """Tests for the :py:class:`collections.Manage` class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.sample = CollectionData(
            id=42,
            name="sid",
            category=CollectionCategory.SUITE,
            full_history_retention_period=7,
            metadata_only_retention_period=14,
            data={"test": 2},
        )
        self.edited = self.sample.model_copy()

    def assertManages(self, *args: str) -> None:
        """Call collection_update with the given arguments match self.edited."""
        command = self.create_command(
            [
                "collection",
                "manage",
                "--workspace=workspace",
                "category",
                "name",
                *args,
            ]
        )
        assert isinstance(command, DebusineCommand)

        with (
            mock.patch.object(
                command.debusine, "collection_get", return_value=self.sample
            ),
            mock.patch.object(
                command.debusine, "collection_update", return_value=self.edited
            ) as collection_update,
            mock.patch.object(command, "show") as show,
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        collection_update.assert_called_once_with("workspace", self.edited)
        show.assert_called_once_with(self.edited)

    def test_rename(self) -> None:
        self.edited.name = "renamed"
        self.assertManages("--rename=renamed")

    def test_set_full_history_retention_period(self) -> None:
        self.edited.full_history_retention_period = 1
        self.assertManages("--full-history-retention-period=1")

    def test_unset_full_history_retention_period(self) -> None:
        self.edited.full_history_retention_period = None
        self.assertManages("--full-history-retention-period=-")

    def test_set_metadata_only_retention_period(self) -> None:
        self.edited.metadata_only_retention_period = 1
        self.assertManages("--metadata-only-retention-period=1")

    def test_unset_metadata_only_retention_period(self) -> None:
        self.edited.metadata_only_retention_period = None
        self.assertManages("--metadata-only-retention-period=-")

    def test_set_data(self) -> None:
        infile = self.create_temporary_file()
        infile.write_text("{'test': 7}")
        self.edited.data = {"test": 7}
        self.assertManages(f"--data={infile}")

    def test_set_data_stdin(self) -> None:
        self.enterContext(self.patch_sys_stdin_read("{'test': 7}"))
        self.edited.data = {"test": 7}
        self.assertManages("--data=-")

    def test_invalid_period(self) -> None:
        stderr, _ = self.capture_output(
            partial(
                self.create_command,
                [
                    "collection",
                    "manage",
                    "--workspace=workspace",
                    "category",
                    "name",
                    "--full-history-retention-period=invalid",
                ],
            ),
            assert_system_exit_code=2,
        )
        self.assertRegex(
            stderr,
            r"error: retention period must be a number or '-'",
        )


class DeleteTests(BaseCliTests):
    """Tests for the :py:class:`collections.Delete` class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.sample = CollectionData(
            id=42,
            name="sid",
            category=CollectionCategory.SUITE,
            full_history_retention_period=7,
            metadata_only_retention_period=14,
            data={"test": 2},
            stats=[],
        )

    def test_delete(self, *args: str) -> None:
        command = self.create_command(
            [
                "collection",
                "delete",
                "--workspace=workspace",
                "category",
                "name",
                *args,
            ]
        )
        assert isinstance(command, DebusineCommand)

        with (
            mock.patch.object(
                command.debusine, "collection_get", return_value=self.sample
            ),
            mock.patch.object(
                command.debusine,
                "collection_delete",
            ) as collection_delete,
            mock.patch.object(command, "feedback") as feedback,
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        collection_delete.assert_called_once_with("workspace", self.sample)
        feedback.assert_called_once_with("Collection deleted.")

    def test_delete_nonempty(self, *args: str) -> None:
        assert self.sample.stats is not None
        self.sample.stats = [
            {
                "type": "artifact",
                "category": "debusine:test",
                "active": True,
                "count": 1,
            },
            {
                "type": "artifact",
                "category": "debusine:test",
                "active": False,
                "count": 15,
            },
        ]
        command = self.create_command(
            [
                "collection",
                "delete",
                "--workspace=workspace",
                "category",
                "name",
                *args,
            ]
        )
        assert isinstance(command, DebusineCommand)

        with (
            mock.patch.object(
                command.debusine, "collection_get", return_value=self.sample
            ),
            mock.patch.object(
                command.debusine,
                "collection_delete",
            ) as collection_delete,
            mock.patch.object(command, "feedback") as feedback,
        ):
            stderr, stdout = self.capture_output(
                command.run, assert_system_exit_code=3
            )

        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr,
            "Collection is not empty"
            " (active debusine:test artifact: 1,"
            " deleted debusine:test artifact: 15):"
            " use --force to delete it.\n",
        )
        collection_delete.assert_not_called()
        feedback.assert_not_called()

    def test_delete_nonempty_force(self, *args: str) -> None:
        assert self.sample.stats is not None
        self.sample.stats += [
            {
                "type": "artifact",
                "category": "test",
                "active": True,
                "count": 1,
            },
            {
                "type": "artifact",
                "category": "test",
                "active": False,
                "count": 15,
            },
        ]
        command = self.create_command(
            [
                "collection",
                "delete",
                "--workspace=workspace",
                "category",
                "name",
                "--force",
                *args,
            ]
        )
        assert isinstance(command, DebusineCommand)

        with (
            mock.patch.object(
                command.debusine, "collection_get", return_value=self.sample
            ),
            mock.patch.object(
                command.debusine,
                "collection_delete",
            ) as collection_delete,
            mock.patch.object(command, "feedback") as feedback,
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        collection_delete.assert_called_once_with("workspace", self.sample)
        feedback.assert_called_once_with("Collection deleted.")


class LookupSingleTests(BaseCliTests):
    """Tests for the :py:class:`collections.LookupSingle` class."""

    def test_no_result(self) -> None:
        command = self.create_command(
            [
                "collection",
                "lookup-single",
                "--workspace",
                "workspace",
                "sid@debian:suite/doesntexist",
            ]
        )
        assert isinstance(command, DebusineCommand)
        detail = "'sid@debian:suite' has no item 'name:doesntexist'"
        with mock.patch.object(
            command.debusine,
            "lookup_single",
            side_effect=DebusineError(
                title="No matches",
                detail=detail,
            ),
        ) as lookup_single:
            stderr, stdout = self.capture_output(
                command.run, assert_system_exit_code=3
            )

        self.assertEqual(stdout, "")
        self.assertEqual(
            yaml.safe_load(stderr),
            {
                "result": "failure",
                "error": {
                    "title": "No matches",
                    "detail": detail,
                },
            },
        )
        lookup_single.assert_called_once_with(
            "sid@debian:suite/doesntexist",
            None,
            'workspace',
            LookupChildType.ANY,
            None,
        )

    def test_success(self) -> None:
        command = self.create_command(
            [
                "collection",
                "lookup-single",
                "--workspace",
                "workspace",
                "sid@debian:suite/libfoo_1.0.0-1",
            ]
        )
        assert isinstance(command, DebusineCommand)
        collection_item = CollectionItemResponse(
            name="libfoo_1.0.0-1",
            data={
                "package": "libfoo",
                "version": "1.0.0-1",
                "component": "main",
                "section": "libs",
            },
        )
        with (
            mock.patch.object(
                command.debusine,
                "lookup_single",
                return_value=LookupSingleResponse(
                    result_type=LookupResultType.ARTIFACT,
                    collection_item=collection_item,
                    artifact=42,
                    collection=None,
                    parent_collection_lookup="sid@debian:suite",
                    links={
                        "self": "/api/a/42",
                        "webui_self": "/a/42",
                        "parent_collection": "/api/c/sid",
                        "webui_parent_collection": "/c/sid",
                    },
                ),
            ) as lookup_single,
            mock.patch(
                "debusine.client.commands.collections.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.collections.rich.print"
            ) as rich_print,
            mock.patch("rich.json.JSON.from_data", side_effect=lambda x: x),
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")

        lookup_single.assert_called_once_with(
            "sid@debian:suite/libfoo_1.0.0-1",
            None,
            'workspace',
            LookupChildType.ANY,
            None,
        )
        rich_print.assert_called_once()
        self.assertEqual(table.add_row.call_count, 1)
        call = table.add_row.call_args_list[0]
        self.assertEqual(call.args[0], "ARTIFACT")
        self.assertEqual(
            call.args[1], "[link=https://debusine.debian.org/a/42]42[/]"
        )
        self.assertEqual(call.args[2], "")
        self.assertEqual(
            call.args[3],
            "[link=https://debusine.debian.org/c/sid]sid@debian:suite[/]",
        )
        self.assertEqual(call.args[4], collection_item.model_dump())

    def test_success_yaml(self) -> None:
        command = self.create_command(
            [
                "collection",
                "lookup-single",
                "--workspace",
                "workspace",
                "sid@debian:suite/libfoo_1.0.0-1",
                "--yaml",
            ]
        )
        assert isinstance(command, DebusineCommand)
        response = LookupSingleResponse(
            result_type=LookupResultType.ARTIFACT,
            collection_item=CollectionItemResponse(
                name="libfoo_1.0.0-1",
                data={
                    "package": "libfoo",
                    "version": "1.0.0-1",
                    "component": "main",
                    "section": "libs",
                },
            ),
            artifact=42,
            collection=None,
            parent_collection_lookup="sid@debian:suite",
        )
        with mock.patch.object(
            command.debusine, "lookup_single", return_value=response
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stderr, "")
        self.assertEqual(
            yaml.safe_load(stdout),
            response.model_dump(exclude_none=True, exclude_unset=True),
        )

    def test_collection_result(self) -> None:
        command = self.create_command(
            [
                "collection",
                "lookup-single",
                "--workspace",
                "workspace",
                "sid@debian:suite",
            ]
        )
        assert isinstance(command, DebusineCommand)
        with (
            # On trixie: https://github.com/coveragepy/coveragepy/issues/1880
            # pragma: no cover
            mock.patch.object(
                command.debusine,
                "lookup_single",
                return_value=LookupSingleResponse(
                    result_type=LookupResultType.COLLECTION,
                    collection_item=None,
                    artifact=None,
                    collection=42,
                    parent_collection_lookup=None,
                    links={
                        "self": "/api/c/42",
                        "webui_self": "/c/42",
                    },
                ),
            ) as lookup_single,
            mock.patch(
                "debusine.client.commands.collections.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.collections.rich.print"
            ) as rich_print,
            mock.patch("rich.json.JSON.from_data", side_effect=lambda x: x),
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")

        lookup_single.assert_called_once_with(
            "sid@debian:suite",
            None,
            'workspace',
            LookupChildType.ANY,
            None,
        )
        rich_print.assert_called_once()
        self.assertEqual(table.add_row.call_count, 1)
        call = table.add_row.call_args_list[0]
        self.assertEqual(call.args[0], "COLLECTION")
        self.assertEqual(call.args[1], "")
        self.assertEqual(
            call.args[2], "[link=https://debusine.debian.org/c/42]42[/]"
        )
        self.assertEqual(call.args[3], "")
        self.assertEqual(call.args[4], "")


class LookupMultipleTests(BaseCliTests):
    """Tests for the :py:class:`collections.LookupMultiple` class."""

    def test_no_result(self) -> None:
        command = self.create_command(
            [
                "collection",
                "lookup-multiple",
                "--workspace",
                "workspace",
                "sid@debian:suite/doesntexist",
            ]
        )
        assert isinstance(command, DebusineCommand)
        detail = "'sid@debian:suite' has no item 'name:doesntexist'"
        with mock.patch.object(
            command.debusine,
            "lookup_multiple",
            side_effect=DebusineError(
                title="No matches",
                detail=detail,
            ),
        ) as lookup_multiple:
            stderr, stdout = self.capture_output(
                command.run, assert_system_exit_code=3
            )

        self.assertEqual(stdout, "")
        self.assertEqual(
            yaml.safe_load(stderr),
            {
                "result": "failure",
                "error": {
                    "title": "No matches",
                    "detail": detail,
                },
            },
        )
        lookup_multiple.assert_called_once_with(
            ["sid@debian:suite/doesntexist"],
            None,
            'workspace',
            LookupChildType.ANY,
            None,
        )

    def test_success(self) -> None:
        command = self.create_command(
            [
                "collection",
                "lookup-multiple",
                "--workspace",
                "workspace",
                "sid@debian:suite/libfoo_1.0.0-1",
                "sid@debian:suite/libbar_1.0.0-1",
            ]
        )
        assert isinstance(command, DebusineCommand)
        collection_items = [
            CollectionItemResponse(
                name="libfoo_1.0.0-1",
                data={
                    "package": "libfoo",
                    "version": "1.0.0-1",
                    "component": "main",
                    "section": "libs",
                },
            ),
            CollectionItemResponse(
                name="libbar_1.0.0-1",
                data={
                    "package": "libbar",
                    "version": "1.0.0-1",
                    "component": "main",
                    "section": "libs",
                },
            ),
        ]
        response = LookupMultipleResponse(
            [
                LookupSingleResponse(
                    result_type=LookupResultType.ARTIFACT,
                    collection_item=collection_item,
                    artifact=artifact,
                    collection=None,
                    parent_collection_lookup="sid@debian:suite",
                    links={
                        "self": "/api/a/42",
                        "webui_self": "/a/42",
                        "parent_collection": "/api/c/sid",
                        "webui_parent_collection": "/c/sid",
                    },
                )
                for (artifact, collection_item) in enumerate(
                    collection_items, start=42
                )
            ]
        )
        with (
            mock.patch.object(
                command.debusine, "lookup_multiple", return_value=response
            ) as lookup_multiple,
            mock.patch(
                "debusine.client.commands.collections.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.collections.rich.print"
            ) as rich_print,
            mock.patch("rich.json.JSON.from_data", side_effect=lambda x: x),
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")

        lookup_multiple.assert_called_once_with(
            [
                "sid@debian:suite/libfoo_1.0.0-1",
                "sid@debian:suite/libbar_1.0.0-1",
            ],
            None,
            'workspace',
            LookupChildType.ANY,
            None,
        )
        rich_print.assert_called_once()
        self.assertEqual(table.add_row.call_count, 2)
        call = table.add_row.call_args_list[0]
        self.assertEqual(call.args[0], "ARTIFACT")
        self.assertEqual(
            call.args[1], "[link=https://debusine.debian.org/a/42]42[/]"
        )
        self.assertEqual(call.args[2], "")
        self.assertEqual(
            call.args[3],
            "[link=https://debusine.debian.org/c/sid]sid@debian:suite[/]",
        )
        self.assertEqual(call.args[4], collection_items[0].model_dump())

    def test_success_yaml(self) -> None:
        command = self.create_command(
            [
                "collection",
                "lookup-multiple",
                "--workspace",
                "workspace",
                "sid@debian:suite/libfoo_1.0.0-1",
                "--yaml",
            ]
        )
        assert isinstance(command, DebusineCommand)
        response = LookupSingleResponse(
            result_type=LookupResultType.ARTIFACT,
            collection_item=CollectionItemResponse(
                name="libfoo_1.0.0-1",
                data={
                    "package": "libfoo",
                    "version": "1.0.0-1",
                    "component": "main",
                    "section": "libs",
                },
            ),
            artifact=42,
            collection=None,
            parent_collection_lookup="sid@debian:suite",
        )
        with mock.patch.object(
            command.debusine, "lookup_multiple", return_value=response
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stderr, "")
        self.assertEqual(
            yaml.safe_load(stdout),
            response.model_dump(exclude_none=True, exclude_unset=True),
        )

    def test_invalid_yaml_lookup_without_key(self) -> None:
        self.enterContext(self.patch_sys_stdin_read(""))
        stderr_writer = self.patch_sys_stderr_write()
        with self.assertRaises(SystemExit):
            self.create_command(
                [
                    "collection",
                    "lookup-multiple",
                    "--workspace",
                    "workspace",
                ]
            )
        stderr_writer.assert_called_once_with(
            "Key 'lookups' not found in YAML input.\n"
        )

    def test_invalid_yaml_lookup_wrong_type(self) -> None:
        self.enterContext(self.patch_sys_stdin_read("lookups: {}"))
        stderr_writer = self.patch_sys_stderr_write()
        with self.assertRaises(SystemExit):
            self.create_command(
                [
                    "collection",
                    "lookup-multiple",
                    "--workspace",
                    "workspace",
                ]
            )
        stderr_writer.assert_called_once_with(
            "Key 'lookups' must be a list of lookups.\n"
        )

    def test_lookups_on_stdin(self) -> None:
        self.enterContext(
            self.patch_sys_stdin_read("lookups: [sid@debian:suite/doesntexist]")
        )
        command = self.create_command(
            [
                "collection",
                "lookup-multiple",
                "--workspace",
                "workspace",
            ]
        )
        assert isinstance(command, DebusineCommand)
        with mock.patch.object(
            command.debusine,
            "lookup_multiple",
            side_effect=DebusineError(
                title="No matches",
            ),
        ) as lookup_multiple:
            self.capture_output(command.run, assert_system_exit_code=3)

        lookup_multiple.assert_called_once_with(
            ["sid@debian:suite/doesntexist"],
            None,
            'workspace',
            LookupChildType.ANY,
            None,
        )

    def test_lookups_from_file(self) -> None:
        infile = self.create_temporary_file()
        infile.write_text("lookups: [sid@debian:suite/doesntexist]")
        command = self.create_command(
            [
                "collection",
                "lookup-multiple",
                "--workspace",
                "workspace",
                "--data",
                str(infile),
            ]
        )
        assert isinstance(command, DebusineCommand)
        with mock.patch.object(
            command.debusine,
            "lookup_multiple",
            side_effect=DebusineError(
                title="No matches",
            ),
        ) as lookup_multiple:
            self.capture_output(command.run, assert_system_exit_code=3)

        lookup_multiple.assert_called_once_with(
            ["sid@debian:suite/doesntexist"],
            None,
            'workspace',
            LookupChildType.ANY,
            None,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for thepage navigation widget."""

from typing import override

import lxml
from django.template import Context

from debusine.db.models import User
from debusine.web.views.table import Column, Table
from debusine.web.views.table.tests.test_table import TableTestCase


class PageNavigationTests(TableTestCase):
    """Tests for :py:class:`PageNavigation`."""

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        for idx in range(10):
            cls.playground.create_user(f"user{idx:02d}")

    @override
    def get_table_class(self) -> type[Table[User]]:
        class _Table(Table[User]):
            user = Column("User", ordering="username")
            default_order = "user"

        return _Table

    def get_page_numbers(
        self, tree: lxml.objectify.ObjectifiedElement
    ) -> list[str]:
        """Return the list of page numbers in the rendered widget."""
        items = tree.xpath("//li[contains(@class, 'page-item')]")
        return [self.get_node_text_normalized(li) for li in items]

    def test_render_single_page(self) -> None:
        """Test rendering when no pagination is needed."""
        table = self._table()
        paginator = table.get_paginator(per_page=20)
        self.assertEqual(paginator.num_pages, 1)
        nav = paginator.page_navigation
        self.assertIs(nav.paginator, paginator)
        self.assertEqual(nav.render(Context()), "")

    def test_render(self) -> None:
        """Test a simple use case."""
        table = self._table()
        paginator = table.get_paginator(per_page=3)
        nav = paginator.page_navigation
        self.assertIs(nav.table, table)
        self.assertIs(nav.paginator, paginator)
        tree = self.assertHTMLValid(nav.render(Context()))
        current = self.assertHasElement(tree, "//li[@class='page-item active']")
        self.assertTextContentEqual(current, "1")
        self.assertEqual(self.get_page_numbers(tree), ["1", "2", "3", "4"])

    def test_ellipsis(self) -> None:
        """Test page navigation with ellipses."""
        for idx in range(20):
            self.playground.create_user(f"user1{idx:02d}")
        table = self._table()
        paginator = table.get_paginator(per_page=3)
        nav = paginator.page_navigation
        self.assertIs(nav.paginator, paginator)
        tree = self.assertHTMLValid(nav.render(Context()))
        self.assertEqual(
            self.get_page_numbers(tree), ["1", "2", "3", "4", "…", "10", "11"]
        )


class PageNavigationPreviewTests(TableTestCase):
    """Tests for :py:class:`PageNavigationPreview`."""

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        for idx in range(10):
            cls.playground.create_user(f"user{idx:02d}")

    @override
    def get_table_class(self) -> type[Table[User]]:
        class _Table(Table[User]):
            user = Column("User", ordering="username")
            default_order = "user"

        return _Table

    def test_render_single_page(self) -> None:
        """Test rendering when no pagination is needed."""
        table = self._table(preview=True)
        paginator = table.get_paginator(per_page=20)
        self.assertEqual(paginator.num_pages, 1)
        nav = paginator.page_navigation
        self.assertIs(nav.paginator, paginator)
        self.assertEqual(nav.render(Context()), "")

    def test_render(self) -> None:
        """Test a simple use case."""
        table = self._table(preview=True)
        paginator = table.get_paginator(per_page=3)
        nav = paginator.page_navigation
        self.assertIs(nav.table, table)
        self.assertIs(nav.paginator, paginator)
        tree = self.assertHTMLValid(nav.render(Context()))
        self.assertTextContentEqual(tree, "3 out of 11 shown")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Utility functions for tests."""

import contextlib
import copy
import datetime as dt
import sys
import traceback
from collections.abc import Callable, Generator, Iterator
from itertools import count
from typing import Any
from urllib.parse import quote, urljoin

from debusine.artifacts.models import DebianSystemTarball
from debusine.client.models import (
    ArtifactResponse,
    FileResponse,
    FilesResponseType,
    PaginatedResponse,
    RemoteArtifact,
    StrictBaseModel,
    WorkRequestResponse,
)
from debusine.tasks import BaseExternalTask

_data_generator_counter = count(0)


def data_generator(size: int) -> Iterator[bytes]:
    """
    Return a data generator of size bytes.

    Each time generates different data.
    """
    while True:
        chunk = bytearray()

        while len(chunk) < size:
            s = str(next(_data_generator_counter)).encode("ascii")
            remaining_space = size - len(chunk)
            chunk.extend(s[:remaining_space])
        yield bytes(chunk)


def yesterday() -> dt.datetime:
    """Return datetime of yesterday."""
    return dt.datetime.now(dt.UTC) - dt.timedelta(days=1)


def tomorrow() -> dt.datetime:
    """Return datetime of tomorrow."""
    return dt.datetime.now(dt.UTC) + dt.timedelta(days=1)


def date_time_to_isoformat_rest_framework(value: dt.datetime) -> str:
    """
    Django REST serialize datetime in a specific way.

    It is reimplemented (simplified) here for Debusine client tests (which
    do not use Django REST).

    See rest_framework/fields.py, DateTimeField.to_representation.
    """
    value_str = value.isoformat()
    if value_str.endswith("+00:00"):
        value_str = value_str[:-6] + "Z"

    return value_str


def create_file_response(**kwargs: Any) -> FileResponse:
    """Return a FileResponse. Use defaults for certain fields."""
    defaults: dict[str, Any] = {
        "size": 0,
        "checksums": {"sha256": "not-used"},
        "type": "file",
        "url": "https://example.com/some-path",
    }
    defaults.update(kwargs)

    return FileResponse(**defaults)


def create_artifact_response(
    scope: str = "debusine",
    **kwargs: Any,
) -> ArtifactResponse:
    """Return an ArtifactResponse. Use defaults for certain fields."""
    kwargs = kwargs.copy()
    defaults: dict[str, Any] = {
        "category": "Testing",
        "data": {},
        "files_to_upload": [],
        "created_at": dt.datetime.now(dt.UTC),
        "scope": scope,
        "workspace": "Testing",
        "download_tar_gz_url": "https://example.com/some-path",
        "files": FilesResponseType(kwargs.pop("files", {})),
    }
    defaults.update(kwargs)
    defaults.setdefault(
        "url",
        f"{quote(scope)}/{quote(defaults['workspace'])}/"
        f"artifact/{defaults['id']}/",
    )

    return ArtifactResponse(**defaults)


def create_remote_artifact(
    scope: str = "debusine",
    **kwargs: Any,
) -> RemoteArtifact:
    """Return a RemoteArtifact. Use defaults for certain fields."""
    defaults: dict[str, Any] = kwargs.copy()
    defaults.setdefault(
        "url",
        f"{quote(scope)}/{quote(defaults['workspace'])}/"
        f"artifact/{defaults['id']}/",
    )

    return RemoteArtifact(**defaults)


def create_work_request_response(
    base_url: str = "https://example.com/",
    scope: str = "debusine",
    **kwargs: Any,
) -> WorkRequestResponse:
    """Return a WorkRequestResponse. Use defaults for certain fields."""
    defaults: dict[str, Any] = {
        "id": 11,
        "parent": 10,
        "workflow_root": 10,
        "created_at": dt.datetime.now(dt.UTC),
        "status": "pending",
        "result": "",
        "task_type": "Worker",
        "task_name": "sbuild",
        "task_data": {},
        "priority_base": 0,
        "priority_adjustment": 0,
        "artifacts": [],
        "scope": "testscope",
        "workspace": "Testing",
        "created_by": 99,
        "aborted_by": None,
    }
    defaults.update(kwargs)

    def wr_webui_url(id_: int) -> str:
        return urljoin(
            base_url,
            f"{quote(scope)}/{quote(defaults['workspace'])}/work-request/{id_}/",
        )

    def wr_api_url(id_: int) -> str:
        return urljoin(base_url, f"api/1.0/work-request/{id_}/")

    defaults.setdefault("url", wr_webui_url(defaults['id']))
    defaults.setdefault("links", {})
    defaults["links"].setdefault("self", wr_api_url(defaults['id']))
    defaults["links"].setdefault("webui_self", wr_webui_url(defaults['id']))
    defaults["links"].setdefault("webui_created_by", "/-/users/123/")
    if defaults["aborted_by"]:
        defaults["links"].setdefault("webui_aborted_by", "/-/users/123/")
    if defaults['parent']:
        defaults["links"].setdefault("parent", wr_api_url(defaults['parent']))
        defaults["links"].setdefault(
            "webui_parent", wr_webui_url(defaults['parent'])
        )
    if defaults['workflow_root']:
        defaults["links"].setdefault(
            "workflow_root", wr_api_url(defaults['workflow_root'])
        )
        defaults["links"].setdefault(
            "webui_workflow_root", wr_webui_url(defaults['workflow_root'])
        )
    return WorkRequestResponse(**defaults)


def create_listing_response(
    *args: StrictBaseModel, next_url: str | None = None
) -> PaginatedResponse:
    """Return a single page of PaginatedResponse, containing args."""
    return PaginatedResponse(
        count=None,
        next=next_url,
        previous=None,
        results=[arg.model_dump() for arg in args],
    )


def create_system_tarball_data(
    codename: str = "sid",
    variant: str | None = None,
    architecture: str = "amd64",
) -> DebianSystemTarball:
    """Return data suitable for a test debian:system-tarball artifact."""
    return DebianSystemTarball(
        filename="system.tar.xz",
        vendor="debian",
        codename=codename,
        components=["main"],
        mirror="https://deb.debian.org/debian",
        variant=variant,
        pkglist={},
        architecture=architecture,
        with_dev=False,
        with_init=False,
    )


@contextlib.contextmanager
def preserve_task_registry() -> Generator[None]:
    """Make task registry changes ephemeral."""
    orig = BaseExternalTask._sub_tasks
    BaseExternalTask._sub_tasks = copy.deepcopy(BaseExternalTask._sub_tasks)
    try:
        yield
    finally:
        BaseExternalTask._sub_tasks = orig


@contextlib.contextmanager
def print_sql_and_stack() -> Generator[None]:
    """Print each SQL query executed, with python stack trace."""
    from django.db import connection

    class Printer:
        """Print numbered SQL queries and stack traces."""

        def __init__(self) -> None:
            """Set the initial counter."""
            self.index = 0

        def __call__(
            self,
            execute: Callable[[str, Any, bool, dict[str, Any]], Any],
            sql: str,
            params: Any,
            many: bool,
            context: dict[str, Any],
        ) -> Any:
            """Interface with the django query executor to print traces."""
            self.index += 1
            print(f"#{self.index}:", "-" * 10, "8<", "-" * 10, file=sys.stderr)
            traceback.print_stack(file=sys.stderr)
            print("  SQL Query:", sql, file=sys.stderr)
            return execute(sql, params, many, context)

    with connection.execute_wrapper(Printer()):
        yield



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Ontology definition of SystemImageBuild."""

import abc
from typing import override

from debusine.tasks import RunCommandTask
from debusine.tasks.models import BaseDynamicTaskData, SystemImageBuildData


class SystemImageBuild(abc.ABC, RunCommandTask[SystemImageBuildData]):
    """Implement ontology SystemImageBuild."""

    TASK_VERSION = 1

    @override
    def get_subject(self) -> str | None:
        return self.data.bootstrap_repositories[0].suite

    @override
    def get_configuration_context(self) -> str | None:
        return self.data.bootstrap_options.architecture

    @override
    def compute_dynamic_data(self) -> BaseDynamicTaskData:
        """Compute dynamic data."""
        suite_name = self.data.bootstrap_repositories[0].suite

        architecture = self.data.bootstrap_options.architecture
        variant = self.data.bootstrap_options.variant
        extra_packages = ",".join(
            sorted(self.data.bootstrap_options.extra_packages)
        )

        if variant is None:
            parameter_summary = f"{suite_name} ({architecture})"
        else:
            parameter_summary = f"{suite_name} ({architecture}-{variant})"

        return BaseDynamicTaskData(
            subject=suite_name,
            parameter_summary=parameter_summary,
            runtime_context=f"{architecture}:{variant}:{extra_packages}",
            configuration_context=architecture,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the Delay task."""

from debusine.server.tasks.wait.delay import Delay
from debusine.server.tasks.wait.models import DelayData
from debusine.server.tasks.wait.tests.helpers import WaitTaskTestCase


class DelayTaskTests(WaitTaskTestCase[Delay]):
    """Test the Delay task."""

    SAMPLE_TASK_DATA = {"delay_until": "2024-01-01T12:00:00+00:00"}

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_missing_delay_until(self) -> None:
        """The `delay_until` field is required."""
        self.task_data = {}
        error_msg = r"delay_until\n\s+Field required.*type=missing"
        with self.assertRaisesRegex(ValueError, error_msg):
            DelayData.model_validate(self.task_data)

    def test_no_additional_properties(self) -> None:
        """Assert no additional properties."""
        self.task_data["input"] = {}
        error_msg = "Extra inputs are not permitted"
        with self.assertRaisesRegex(ValueError, error_msg):
            DelayData.model_validate(self.task_data)

    def test_label(self) -> None:
        """Test get_label."""
        self.assertEqual(
            self.task.get_label(), "delay until 2024-01-01 12:00:00+00:00"
        )



# Generated by Django 4.2.21 on 2025-11-03 12:10

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0024_workrequest_db_workrequest_workflow_has_root'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='artifact',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='artifactrelation',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='asset',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='assetusage',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='collection',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='collectionitem',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='file',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='group',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='groupauditlog',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='scope',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='token',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='user',
            options={
                'base_manager_name': 'objects',
                'verbose_name': 'user',
                'verbose_name_plural': 'users',
            },
        ),
        migrations.AlterModelOptions(
            name='worker',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='workerpool',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='workflowtemplate',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='workrequest',
            options={
                'base_manager_name': 'objects',
                'ordering': ['id'],
                'permissions': [
                    (
                        'manage_workrequest_priorities',
                        'Can set positive priority adjustments on work requests',
                    )
                ],
            },
        ),
        migrations.AlterModelOptions(
            name='workspace',
            options={'base_manager_name': 'objects'},
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common infrastructure for testing the debusine Cli class."""

import contextlib
import io
import logging
import signal
from collections.abc import Callable, Generator
from pathlib import Path
from typing import Any, override
from unittest import mock
from unittest.mock import MagicMock
from urllib.parse import urlsplit, urlunsplit

from debusine.client.cli import Cli
from debusine.client.commands.base import Command
from debusine.client.exceptions import DebusineError
from debusine.test import TestCase


class BaseCliTests(TestCase):
    """Basic functionality to implement tests for the Cli class."""

    @override
    def setUp(self) -> None:
        super().setUp()

        debian_server = {
            'api-url': 'https://debusine.debian.org/api',
            'scope': 'debian',
            'token': 'token-for-debian',
        }
        kali_server = {
            'api-url': 'https://debusine.kali.org/api',
            'scope': 'kali',
            'token': 'token-for-kali',
        }

        self.servers = {
            "debian": debian_server,
            "kali": kali_server,
        }
        self.default_server = "debian"

        self.default_sigint_handler = signal.getsignal(signal.SIGINT)
        self.default_sigterm_handler = signal.getsignal(signal.SIGTERM)

    @override
    def tearDown(self) -> None:
        # Restore signal handlers. Cli.execute() changes them
        signal.signal(signal.SIGINT, self.default_sigint_handler)
        signal.signal(signal.SIGTERM, self.default_sigterm_handler)

        # Cli._build_debusine_object and Cli._setup_http_logging reconfigure
        # the logging system.  Reset it.
        for logger in logging.Logger.manager.loggerDict.values():
            if not isinstance(logger, logging.PlaceHolder):
                for handler in list(logger.handlers):
                    logger.removeHandler(handler)
                    handler.close()
                logger.propagate = True

        super().tearDown()

    def create_cli(self, argv: list[str], create_config: bool = True) -> Cli:
        """
        Return a Cli object using argv, self.stdout and self.stderr.

        :param argv: arguments passed to the Cli class.
        :param create_config: True for creating a config file and adding
          --config config_file_path into CLI's argv.
        """
        if create_config:
            if '--config' in argv:  # pragma: no cover
                raise ValueError(
                    'Incompatible options: create_config cannot be True if '
                    '--config is in argv'
                )
            config = self.create_config_file()

            argv = ['--config', config.as_posix(), *argv]

        return Cli(argv)

    def create_command(
        self, argv: list[str], create_config: bool = True
    ) -> Command:
        """
        Return a Command object instantiated using argv.

        :param argv: arguments passed to the Cli class.
        :param create_config: True for creating a config file and adding
          --config config_file_path into CLI's argv.
        """
        cli = self.create_cli(argv, create_config=create_config)
        return cli.get_command()

    def create_config_file(self) -> Path:
        """Write a config file and returns the path."""
        config_directory = self.create_temp_config_directory(
            {
                'General': {'default-server': self.default_server},
                'server:debian': self.servers["debian"],
                'server:kali': self.servers["kali"],
            }
        )

        return config_directory / 'config.ini'

    def capture_output(
        self,
        func: Callable[..., Any],
        args: list[Any] | None = None,
        assert_system_exit_code: int | None = None,
    ) -> tuple[str, str]:
        """
        Execute func() and return stderr and stdout output.

        :param func: functor to be executed
        :param args: list of arguments to be passed to func() (or None if
          no arguments are passed when calling func()
        :param assert_system_exit_code: if not None assert that SystemExit
          is raised
        :return: tuple[stderr, stdout]
        """
        if args is None:
            args = []

        stderr = io.StringIO()
        stdout = io.StringIO()

        with (
            contextlib.redirect_stderr(stderr),
            contextlib.redirect_stdout(stdout),
        ):
            if assert_system_exit_code is not None:
                with self.assertRaisesSystemExit(assert_system_exit_code):
                    func(*args)
            else:
                func(*args)

        return stderr.getvalue(), stdout.getvalue()

    @contextlib.contextmanager
    def patch_sys_stdin_read(self, data: str) -> Generator[mock.MagicMock]:
        """Patch sys.stdin.read to return what a user might write / input."""
        with mock.patch('sys.stdin') as mock_stdin:
            mock_stdin.read.return_value = data
            yield mock_stdin

    def patch_sys_stderr_write(self) -> MagicMock:
        """Patch sys.stderr.write to check what was written."""
        patcher_sys_stderr = mock.patch("sys.stderr.write")
        mocked_sys_stderr = patcher_sys_stderr.start()
        self.addCleanup(patcher_sys_stderr.stop)

        return mocked_sys_stderr

    def get_base_url(self, server_name: str) -> str:
        """Return the base web URL for a given server."""
        split_server = urlsplit(self.servers[server_name]["api-url"])
        return urlunsplit(
            (split_server.scheme, split_server.netloc, "", "", "")
        )

    def assertDebusineError(
        self, exception: Exception, data: dict[str, Any]
    ) -> None:
        """Ensure the exception is a Debusine error with the given data."""
        assert isinstance(exception, DebusineError)
        self.assertEqual(exception.asdict(), data)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Template tag library for workflows."""

from django import template
from django.utils.html import format_html
from django.utils.safestring import SafeString

from debusine.artifacts.models import TaskTypes
from debusine.db.models import WorkRequest

register = template.Library()


@register.simple_tag(takes_context=True)
def workflow_runtime_status_small(
    context: template.Context, workflow: WorkRequest
) -> str:
    """
    Render workflow runtime status component.

    It displays a badge with one letter for the workflow.status. If the
    status is RUNNING, it displays a second batch with
    workflow.workflow_runtime_status.
    """
    statuses_formatting: dict[WorkRequest.Statuses, tuple[str, str]] = {
        WorkRequest.Statuses.PENDING: ("secondary", "P"),
        WorkRequest.Statuses.RUNNING: ("info", "R"),
        WorkRequest.Statuses.COMPLETED: ("primary", "C"),
        WorkRequest.Statuses.ABORTED: ("dark", "A"),
        WorkRequest.Statuses.BLOCKED: ("dark", "B"),
    }

    runtime_status_formatting: dict[
        WorkRequest.RuntimeStatuses, tuple[str, str]
    ] = {
        WorkRequest.RuntimeStatuses.NEEDS_INPUT: ("info", "I"),
        WorkRequest.RuntimeStatuses.RUNNING: ("info", "R"),
        WorkRequest.RuntimeStatuses.WAITING: ("info", "W"),
        WorkRequest.RuntimeStatuses.PENDING: ("secondary", "P"),
        WorkRequest.RuntimeStatuses.ABORTED: ("dark", "A"),
        WorkRequest.RuntimeStatuses.COMPLETED: ("primary", "C"),
        WorkRequest.RuntimeStatuses.BLOCKED: ("dark", "B"),
    }

    if (
        workflow.status == WorkRequest.Statuses.RUNNING
        and workflow.workflow_runtime_status is not None
    ):
        # Return badge with the workflow.workflow_runtime_status only
        workflow_runtime_status = WorkRequest.RuntimeStatuses(
            workflow.workflow_runtime_status
        )

        runtime_status_color, runtime_status_text = runtime_status_formatting[
            workflow_runtime_status
        ]
        runtime_status_title = workflow_runtime_status.label

        runtime_status_html = format_html(
            '<span title="{runtime_status_title}" '
            'class="badge text-bg-{runtime_status_color}">'
            '{runtime_status_text}</span>',
            runtime_status_title=runtime_status_title,
            runtime_status_color=runtime_status_color,
            runtime_status_text=runtime_status_text,
        )

        if workflow_runtime_status == WorkRequest.RuntimeStatuses.NEEDS_INPUT:
            work_request_needs_input = _first_workrequest_needs_input(
                context["workflows_flattened"][workflow.id]
            )
            if work_request_needs_input is not None:
                runtime_status_html = format_html(
                    '<a href="{url}">{runtime_status_html}</a>',
                    url=work_request_needs_input.get_absolute_url(),
                    runtime_status_html=runtime_status_html,
                )

        runtime_status_html = SafeString(" " + runtime_status_html)

    else:
        runtime_status_html = SafeString("")

    status = WorkRequest.Statuses(workflow.status)

    status_color, status_text = statuses_formatting[status]

    status_title = status.label

    html = format_html(
        '<span title="{status_title}" class="badge text-bg-{status_color}">'
        '{status_text}</span>'
        '{runtime_status_html}',
        status_title=status_title,
        status_color=status_color,
        status_text=status_text,
        runtime_status_html=runtime_status_html,
    )

    return html


def _first_workrequest_needs_input(
    workflow_flattened: list[WorkRequest],
) -> WorkRequest | None:
    for work_request in sorted(
        workflow_flattened, key=lambda workrequest: workrequest.created_at
    ):
        if (
            work_request.task_type == TaskTypes.WAIT
            and work_request.status == WorkRequest.Statuses.RUNNING
            and work_request.workflow_data_json.get("needs_input")
        ):
            return work_request

    return None



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Database interaction for worker tasks on the server.

This DBFieldResolver extension is used when Task code is run server-side with
database access.
"""

from collections.abc import Collection as AbcCollection
from typing import Any, TYPE_CHECKING, assert_never, overload, override

from django.conf import settings

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.client.models import LookupChildType, RelationType
from debusine.db.models.artifacts import Artifact
from debusine.db.models.collections import Collection
from debusine.db.permissioncontext import PermissionContext
from debusine.db.tasks.models import (
    DBArtifactInfo,
    DBBaseTaskInputArtifact,
    DBInputArtifactMultiple,
    DBInputCollectionSingle,
)
from debusine.tasks import (
    BaseTaskWithExecutor,
    TaskConfigError,
    ensure_collection_category,
)
from debusine.tasks.executors import ExecutorImageCategory
from debusine.tasks.inputs import inputs
from debusine.tasks.inputs.resolver import FieldResolver
from debusine.tasks.models import (
    BackendType,
    ExtraDebusineRepository,
    ExtraExternalRepository,
    ExtraRepository,
    InputCollectionSingle,
    LookupMultiple,
    LookupSingle,
)

if TYPE_CHECKING:
    from debusine.db.tasks import DBTask


class DBFieldResolver(FieldResolver):
    """Implementation of database interaction in worker tasks."""

    def __init__(self, task: "DBTask[Any, Any]") -> None:
        """Construct a :py:class:`DBFieldResolver`."""
        self.task = task
        self.work_request = task.work_request

    @overload
    def lookup_single_artifact(
        self,
        lookup: LookupSingle,
        *,
        default_category: CollectionCategory | None = None,
        label: str,
    ) -> DBArtifactInfo: ...

    @overload
    def lookup_single_artifact(
        self,
        lookup: None,
        *,
        default_category: CollectionCategory | None = None,
        label: str,
    ) -> None: ...

    @override
    def lookup_single_artifact(
        self,
        lookup: LookupSingle | None,
        *,
        default_category: CollectionCategory | None = None,
        label: str,
    ) -> DBArtifactInfo | None:
        return (
            None
            if lookup is None
            else DBArtifactInfo.from_artifact(
                self.work_request.lookup_single(
                    lookup=lookup,
                    default_category=default_category,
                    expect_type=LookupChildType.ARTIFACT,
                ).artifact,
                label=label,
                lookup=lookup,
            )
        )

    @override
    def lookup_multiple_artifacts(
        self,
        lookup: LookupMultiple,
        *,
        default_category: CollectionCategory | None = None,
        label: str,
    ) -> DBInputArtifactMultiple:
        artifacts = [
            DBBaseTaskInputArtifact.from_artifact(result.artifact)
            for result in self.work_request.lookup_multiple(
                lookup=lookup,
                default_category=default_category,
                expect_type=LookupChildType.ARTIFACT,
            )
        ]
        return DBInputArtifactMultiple(
            lookup=lookup, label=label, artifacts=artifacts
        )

    @override
    def find_related_artifacts(
        self,
        artifact_ids: AbcCollection[int],
        *,
        target_category: ArtifactCategory,
        relation_type: RelationType = RelationType.RELATES_TO,
        label: str,
    ) -> DBInputArtifactMultiple:
        artifacts = [
            DBBaseTaskInputArtifact.from_artifact(artifact)
            for artifact in Artifact.objects.filter(
                targeted_by__artifact__in=artifact_ids,
                category=target_category,
                targeted_by__type=relation_type,
            ).order_by("id")
        ]
        # TODO: can we do better for a lookup value here?
        return DBInputArtifactMultiple(
            lookup=[], label=label, artifacts=artifacts
        )

    @override
    def get_server_setting(self, setting: str) -> str:
        value = getattr(settings, setting)
        assert isinstance(value, str)
        return value

    @override
    def get_fallback_architecture(self) -> str:
        return self.work_request.get_fallback_architecture()

    @override
    def export_suite_signing_keys(self, suite_id: int) -> str | None:
        # Import here to prevent circular imports
        from debusine.server.collections import DebianSuiteManager

        suite = Collection.objects.get(id=suite_id)
        assert isinstance(suite.manager, DebianSuiteManager)
        return suite.manager.export_signing_keys()

    @override
    def lookup_environment(
        self,
        lookup: LookupSingle,
        *,
        architecture: str | None = None,
        backend: BackendType | None = None,
        default_category: CollectionCategory | None = None,
        image_category: "ExecutorImageCategory | None" = None,
        try_variant: str | None = None,
        label: str,
    ) -> DBArtifactInfo:
        return (
            None
            if lookup is None
            else DBArtifactInfo.from_artifact(
                self.work_request.lookup_environment(
                    lookup,
                    architecture=architecture,
                    backend=backend,
                    default_category=default_category,
                    image_category=image_category,
                    try_variant=try_variant,
                ).artifact,
                lookup=lookup,
                label=label,
            )
        )

    @override
    def resolve_environmentinput(
        self, field: inputs.EnvironmentInput
    ) -> DBArtifactInfo:
        # Prevent import loop
        from debusine.db.tasks import DBWorkerProxyTask

        assert isinstance(self.task, DBWorkerProxyTask)
        external_task = self.task.task
        assert isinstance(external_task, BaseTaskWithExecutor)
        lookup = self.get_field(field)
        assert lookup is not None
        info = self.lookup_environment(
            lookup,
            architecture=external_task.build_architecture,
            backend=external_task.backend if field.set_backend else None,
            default_category=CollectionCategory.ENVIRONMENTS,
            image_category=field.image_category,
            try_variant=self.task.name if field.try_variant else None,
            label=field.field,
        )
        field.check_category(info)
        return info

    @override
    def resolve_optionalenvironmentinput(
        self, field: inputs.OptionalEnvironmentInput
    ) -> DBArtifactInfo | None:
        # Prevent import loop
        from debusine.db.tasks import DBWorkerProxyTask

        assert isinstance(self.task, DBWorkerProxyTask)
        external_task = self.task.task
        assert isinstance(external_task, BaseTaskWithExecutor)
        lookup = self.get_field(field)
        if lookup is None:
            return None
        info = self.lookup_environment(
            lookup,
            architecture=external_task.build_architecture,
            backend=external_task.backend,
            default_category=CollectionCategory.ENVIRONMENTS,
            image_category=None,
            try_variant=self.task.name,
            label=field.field,
        )
        field.check_category(info)
        return info

    @override
    def resolve_suitearchiveinput(
        self, field: inputs.SuiteArchiveInput
    ) -> tuple[DBInputCollectionSingle, DBInputCollectionSingle | None]:
        # Import here to prevent circular imports.
        from debusine.server.collections.lookup import lookup_single

        lookup = self.get_field(field)
        assert lookup is not None

        suite_result = self.work_request.lookup_single(
            lookup=lookup,
            expect_type=LookupChildType.COLLECTION,
        )
        suite = DBInputCollectionSingle.from_collection(
            suite_result.collection, lookup=lookup, label=field.field
        )

        ensure_collection_category(
            configuration_key=field.field,
            category=suite.category,
            expected=CollectionCategory.SUITE,
        )

        archive_lookup = f"_@{CollectionCategory.ARCHIVE}"
        archive: InputCollectionSingle | None
        try:
            archive_result = lookup_single(
                archive_lookup,
                workspace=suite_result.collection.workspace,
                pc=PermissionContext.for_work_request(
                    work_request=self.work_request
                ),
                expect_type=LookupChildType.COLLECTION,
            )
        except KeyError:
            archive = None
        else:
            archive = (
                DBInputCollectionSingle.from_collection(
                    archive_result.collection,
                    lookup=archive_lookup,
                    label=field.field,
                )
                if archive_result is not None
                else None
            )

        return suite, archive

    def _resolve_extra_repository(
        self,
        extra_repository: ExtraRepository,
    ) -> ExtraExternalRepository:
        """Resolve a single extra repository into the external form."""
        # Import here to prevent circular imports
        from debusine.server.collections import DebianSuiteManager

        match extra_repository:
            case ExtraDebusineRepository():
                debusine_fqdn = settings.DEBUSINE_DEBIAN_ARCHIVE_PRIMARY_FQDN
                suite = self.work_request.lookup_single(
                    extra_repository.suite,
                    expect_type=LookupChildType.COLLECTION,
                ).collection
                assert isinstance(suite.manager, DebianSuiteManager)
                components = extra_repository.components
                if components is None:
                    if "components" not in suite.data:
                        raise TaskConfigError(
                            f"'components' not set for {extra_repository.suite}"
                        )
                    components = suite.data["components"]
                return ExtraExternalRepository(
                    # Ideally we'd use HTTPS, but that might require
                    # certificate setup within environments, and we know that
                    # our own repositories are signed, so HTTP should be good
                    # enough to use from workers for now.
                    url=(
                        f"http://{debusine_fqdn}"
                        f"/{suite.workspace.scope.name}/{suite.workspace.name}"
                    ),
                    suite=suite.name,
                    components=components,
                    signing_key=suite.manager.export_signing_keys(),
                )
            case ExtraExternalRepository():
                return extra_repository
            case _ as unreachable:
                assert_never(unreachable)

    @override
    def resolve_extrarepositoriesinput(
        self, field: inputs.ExtraRepositoriesInput
    ) -> list[ExtraExternalRepository] | None:
        extra_repositories = self.get_field(field)
        if extra_repositories is None:
            return None
        return [
            self._resolve_extra_repository(extra_repository)
            for extra_repository in extra_repositories
        ]

    @override
    def get_field[T](self, field: inputs.DataFieldInput[T, Any]) -> T | None:
        try:
            value: Any = field.getter(self.task.data)
        except AttributeError:
            raise TaskConfigError(
                f"Task input attribute {field.field!r} not found"
            )
        if value is None:
            return None
        if not isinstance(value, field.lookup_type):
            raise TaskConfigError(
                f"Task input attribute {field.field!r} is {value!r}"
                f" which is not of type {field.lookup_type}"
            )
        return value



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""URLs related to group management."""

from django.urls import path

import debusine.web.views.auth as views

app_name = "groups"

# Scoped URLs
urlpatterns = [
    path(
        "<group>/",
        views.GroupDetailView.as_view(),
        name="detail",
    ),
    path(
        "<group>/edit/<user>/",
        views.MembershipUpdateView.as_view(),
        name="update-member",
    ),
    path(
        "<group>/remove/<user>/",
        views.MembershipDeleteView.as_view(),
        name="remove-member",
    ),
    path(
        "<group>/audit-log/",
        views.GroupAuditLogView.as_view(),
        name="audit-log",
    ),
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine worker launcher."""

import argparse
import asyncio
import logging
import sys
from collections.abc import Sequence

from debusine.utils import debusine_version
from debusine.worker import Worker
from debusine.worker.config import ConfigHandler


def main(argv: Sequence[str]) -> None:
    """Run the debusine-worker command (main entry-point)."""
    config_paths = ', '.join(ConfigHandler.default_directories)
    parser = argparse.ArgumentParser(
        prog='debusine-worker',
        description=f'Start Debusine worker process: register if needed and '
        f'connects to the server. The configuration is searched in '
        f'the files config.ini and token in the '
        f'directories {config_paths})',
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {debusine_version() or 'uninstalled-version'}",
    )
    parser.add_argument(
        '--log-level',
        help='Minimum log level. Overrides log-level (in [General] section) '
        'from config.ini. If not specified anywhere logs '
        f'{logging.getLevelName(Worker.DEFAULT_LOG_LEVEL)} and above.',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
    )
    parser.add_argument(
        '--working-directory',
        type=str,
        help='Execute tasks within this directory. Overrides working-directory '
        '(in [General] section) from config.ini. If not specified anywhere '
        'executes tasks in $HOME/workdir.',
    )

    args = parser.parse_args(argv)

    async def async_main() -> None:
        async with Worker(
            log_level=args.log_level,
            working_directory=args.working_directory,
        ) as worker:
            await worker.main()

    asyncio.run(async_main())


def entry_point() -> None:  # pragma: no cover
    """Entry point for the packaging."""
    main(sys.argv[1:])


if __name__ == '__main__':
    entry_point()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Rename various workflow data fields to ``qa_suite``."""

from django.db import migrations

from debusine.artifacts.models import TaskTypes
from debusine.db.migrations._utils import (
    make_work_request_task_data_field_renamer,
    make_workflow_template_task_data_field_renamer,
)


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0012_alter_collectionitem_created_by_workflow_and_more"),
    ]
    replaces = [
        ("db", "0154_qa_suite"),
    ]

    operations = [
        make_work_request_task_data_field_renamer(
            TaskTypes.WORKFLOW,
            "reverse_dependencies_autopkgtest",
            [("suite_collection", "qa_suite")],
        ),
        make_workflow_template_task_data_field_renamer(
            "reverse_dependencies_autopkgtest",
            [("suite_collection", "qa_suite")],
        ),
        make_work_request_task_data_field_renamer(
            TaskTypes.WORKFLOW,
            {"qa", "debian_pipeline"},
            [("reverse_dependencies_autopkgtest_suite", "qa_suite")],
        ),
        make_workflow_template_task_data_field_renamer(
            {"qa", "debian_pipeline"},
            [("reverse_dependencies_autopkgtest_suite", "qa_suite")],
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Pydantic models used by debusine workflows."""

import datetime as dt
from enum import StrEnum
from functools import partial
from typing import Annotated, Any, Self

import pydantic

from debusine.assets import KeyPurpose
from debusine.db.models.workspaces import workspace_name_regex
from debusine.server.tasks.models import PackageUploadTarget
from debusine.tasks.models import (
    ActionRetryWithDelays,
    AutopkgtestFailOn,
    AutopkgtestNeedsInternet,
    AutopkgtestTimeout,
    BackendType,
    BaseDynamicTaskData,
    BaseTaskData,
    BlhcFlags,
    DebDiffFlags,
    ExtraRepository,
    LintianFailOnSeverity,
    LintianOutput,
    LookupMultiple,
    LookupSingle,
    SbuildBinNMU,
    SbuildInput,
    require_unique,
)


class BaseWorkflowData(BaseTaskData):
    """
    Base class for workflow data.

    Workflow data is encoded as JSON in the database, and it is modeled as a
    pydantic data structure in memory for both ease of access and validation.
    """


class WorkRequestManualUnblockAction(StrEnum):
    """An action taken on a review of a manual unblock request."""

    ACCEPT = "accept"
    REJECT = "reject"


class WorkRequestManualUnblockLog(pydantic.BaseModel):
    """A log entry for a review of a manual unblock request."""

    user_id: int
    timestamp: dt.datetime
    notes: str | None = None
    action: WorkRequestManualUnblockAction | None = None


class WorkRequestManualUnblockData(pydantic.BaseModel):
    """Data for a manual unblock request."""

    log: list[WorkRequestManualUnblockLog] = pydantic.Field(
        default_factory=list
    )


class WorkRequestWorkflowData(pydantic.BaseModel):
    """Data structure for WorkRequest.workflow_data."""

    model_config = pydantic.ConfigDict(validate_assignment=True, extra="forbid")

    #: Whether the parent workflow and reverse-dependencies should continue
    #: if this work request fails.  If False, they are aborted instead.
    allow_failure: bool = pydantic.Field(default=False, frozen=True)

    #: Whether this work request should continue if any of its child work
    #: requests or dependencies fails.  If False, it is aborted instead.
    allow_dependency_failures: bool = pydantic.Field(default=False, frozen=True)

    #: name of the step in the visual representation of the workflow
    display_name: str | None = pydantic.Field(default=None, frozen=True)

    #: internal identifier used to differentiate multiple workflow callbacks
    #: inside a single workflow. It allows the orchestrator to encode the plan
    #: about what it is supposed to do at this point in the workflow.
    step: str | None = pydantic.Field(default=None, frozen=True)

    #: Name of the group within this workflow containing this work request.
    group: str | None = pydantic.Field(default=None, frozen=True)

    #: For workflows created from a workflow template, the name of that
    #: template, cached here for convenience.
    workflow_template_name: str | None = pydantic.Field(
        default=None, frozen=True
    )

    #: Data stored for manual unblock requests.
    manual_unblock: WorkRequestManualUnblockData | None = None

    #: Whether this task requires user input.  Only relevant for WAIT tasks.
    needs_input: bool | None = None

    #: The number of times this task has been retried automatically.
    retry_count: int = pydantic.Field(default=0)

    #: Whether this task should be shown in the visual representation of the
    #: workflow.
    visible: bool = pydantic.Field(default=True, frozen=True)

    #: Whether this task's descendants should be expanded by default in the
    #: visual representation of the workflow.
    expanded: bool = pydantic.Field(default=True, frozen=True)


class SbuildWorkflowData(BaseWorkflowData):
    """Sbuild workflow data."""

    prefix: str = ""
    input: SbuildInput
    target_distribution: str
    # If AUTO is used, default to BackendType.UNSHARE
    backend: BackendType = BackendType.AUTO
    architectures: Annotated[
        list[str],
        pydantic.Field(min_length=1),
        pydantic.AfterValidator(require_unique),
    ]
    arch_all_build_architecture: str = "amd64"
    environment_variant: str | None = None
    extra_repositories: list[ExtraRepository] | None = None
    binnmu: SbuildBinNMU | None = None
    build_profiles: list[str] | None = None
    retry_delays: Annotated[
        list[
            Annotated[
                str, pydantic.Field(pattern=ActionRetryWithDelays._delay_re)
            ]
        ]
        | None,
        pydantic.Field(min_length=1),
    ] = None
    signing_template_names: dict[str, list[str]] = {}


class SbuildWorkflowDynamicData(BaseDynamicTaskData):
    """Dynamic data for the Sbuild workflow."""

    source_artifact_id: int | None = None


class UpdateEnvironmentsWorkflowTarget(BaseWorkflowData):
    """A target for an update_environments workflow."""

    codenames: str | list[str]
    codename_aliases: dict[str, list[str]] = pydantic.Field(
        default_factory=dict
    )
    variants: str | None | list[str | None] = pydantic.Field(
        default_factory=list
    )
    backends: str | list[str] = pydantic.Field(default_factory=list)
    architectures: Annotated[
        list[str],
        pydantic.Field(min_length=1),
        pydantic.AfterValidator(require_unique),
    ]
    mmdebstrap_template: dict[str, Any] | None = None
    simplesystemimagebuild_template: dict[str, Any] | None = None


class UpdateEnvironmentsWorkflowData(BaseWorkflowData):
    """update_environments workflow data."""

    vendor: str
    targets: Annotated[
        list[UpdateEnvironmentsWorkflowTarget], pydantic.Field(min_length=1)
    ]


class PackageUploadWorkflowData(BaseWorkflowData):
    """`package_upload` workflow data."""

    prefix: str = ""

    source_artifact: LookupSingle | None = None
    binary_artifacts: LookupMultiple = pydantic.Field(
        default_factory=partial(LookupMultiple, ())
    )
    merge_uploads: bool = False
    since_version: str | None = None
    target_distribution: str | None = None
    key: str | None = None
    binary_key: str | None = None
    require_signature: bool = True
    target: PackageUploadTarget
    delayed_days: int | None = None
    vendor: str | None = None
    codename: str | None = None
    arch_all_build_architecture: str = "amd64"


class PackageUploadWorkflowDynamicData(BaseDynamicTaskData):
    """Dynamic data for PackageUpload workflow."""

    source_artifact_id: int | None = None
    binary_artifacts_ids: list[int] | None = None


class MakeSignedSourceWorkflowData(BaseWorkflowData):
    """`make_signed_source` workflow data."""

    prefix: str = ""

    binary_artifacts: LookupMultiple
    signing_template_artifacts: LookupMultiple

    vendor: str
    codename: str
    architectures: list[str]
    purpose: KeyPurpose
    key: str
    sbuild_backend: BackendType = BackendType.AUTO


class MakeSignedSourceWorkflowDynamicData(BaseDynamicTaskData):
    """Dynamic data for the MakeSignedSource workflow."""

    binary_artifacts_ids: list[int] | None = None
    signing_template_artifacts_ids: list[int] | None = None


class RegressionTrackingUpdateQAResults(StrEnum):
    """Whether a regression-tracking workflow should update QA results."""

    NO = "no"
    YES = "yes"
    FORCE = "force"


def _convert_update_qa_results(value: Any) -> RegressionTrackingUpdateQAResults:
    match value:
        case False:
            return RegressionTrackingUpdateQAResults.NO
        case True:
            return RegressionTrackingUpdateQAResults.YES
        case _:
            return RegressionTrackingUpdateQAResults(value)


class RegressionTrackingWorkflowData(BaseWorkflowData):
    """Common data for workflows that handle regression tracking."""

    prefix: str = ""
    reference_prefix: str = ""

    qa_suite: LookupSingle | None = None
    reference_qa_results: LookupSingle | None = None
    enable_regression_tracking: bool = False
    update_qa_results: Annotated[
        RegressionTrackingUpdateQAResults,
        pydantic.BeforeValidator(
            _convert_update_qa_results, json_schema_input_type=str | bool
        ),
    ] = RegressionTrackingUpdateQAResults.NO

    @pydantic.model_validator(mode="after")
    def check_qa_results_consistency(self) -> Self:
        """Check consistency of regression-tracking options."""
        if self.enable_regression_tracking or self.update_qa_results in {
            RegressionTrackingUpdateQAResults.YES,
            RegressionTrackingUpdateQAResults.FORCE,
        }:
            if self.qa_suite is None:
                raise ValueError(
                    '"qa_suite" is required if "enable_regression_tracking" or '
                    '"update_qa_results" is set'
                )
            if self.reference_qa_results is None:
                raise ValueError(
                    '"reference_qa_results" is required if '
                    '"enable_regression_tracking" or "update_qa_results" is set'
                )
        if self.enable_regression_tracking and not self.reference_prefix:
            raise ValueError(
                '"reference_prefix" is required if '
                '"enable_regression_tracking" is set'
            )
        return self


class PiupartsWorkflowData(RegressionTrackingWorkflowData):
    """`piuparts` workflow data."""

    source_artifact: LookupSingle
    binary_artifacts: LookupMultiple

    vendor: str
    codename: str
    architectures: list[str] | None = None
    backend: BackendType = BackendType.AUTO
    environment: LookupSingle | None = None
    extra_repositories: list[ExtraRepository] | None = None


class PiupartsWorkflowDynamicData(BaseDynamicTaskData):
    """Dynamic data for Piuparts workflow."""

    source_artifact_id: int | None = None
    binary_artifacts_ids: list[int] | None = None


class AutopkgtestWorkflowData(RegressionTrackingWorkflowData):
    """`autopkgtest` workflow data."""

    source_artifact: LookupSingle
    binary_artifacts: LookupMultiple
    context_artifacts: LookupMultiple = pydantic.Field(
        default_factory=partial(LookupMultiple, ())
    )

    vendor: str
    codename: str
    backend: BackendType = BackendType.AUTO
    architectures: list[str] = pydantic.Field(default_factory=list)
    arch_all_build_architecture: str = "amd64"
    extra_repositories: list[ExtraRepository] | None = None

    include_tests: list[str] = pydantic.Field(default_factory=list)
    exclude_tests: list[str] = pydantic.Field(default_factory=list)
    debug_level: int = pydantic.Field(default=0, ge=0, le=3)
    pin_packages: dict[str, list[str]] = pydantic.Field(default_factory=dict)
    extra_environment: dict[str, str] = pydantic.Field(default_factory=dict)
    needs_internet: AutopkgtestNeedsInternet = AutopkgtestNeedsInternet.RUN
    fail_on: AutopkgtestFailOn = pydantic.Field(
        default_factory=AutopkgtestFailOn
    )
    timeout: AutopkgtestTimeout | None = None


class AutopkgtestWorkflowDynamicData(BaseDynamicTaskData):
    """Dynamic data for the Autopkgtest workflow."""

    source_artifact_id: int | None = None
    binary_artifacts_ids: list[int] | None = None
    context_artifacts_ids: list[int] | None = None


class ReverseDependenciesAutopkgtestWorkflowData(
    RegressionTrackingWorkflowData
):
    """`reverse_dependencies_autopkgtest` workflow data."""

    source_artifact: LookupSingle
    binary_artifacts: LookupMultiple
    context_artifacts: LookupMultiple = pydantic.Field(
        default_factory=partial(LookupMultiple, ())
    )

    # RegressionTrackingWorkflowData declares this as optional, but it's
    # required here.
    qa_suite: LookupSingle

    vendor: str
    codename: str
    backend: BackendType = BackendType.AUTO
    architectures: list[str] = []
    arch_all_build_architecture: str = "amd64"
    packages_allowlist: list[str] | None = None
    packages_denylist: list[str] = []
    extra_repositories: list[ExtraRepository] | None = None

    debug_level: int = pydantic.Field(default=0, ge=0, le=3)


class ReverseDependenciesAutopkgtestWorkflowDynamicData(BaseDynamicTaskData):
    """Dynamic data for the ReverseDependenciesAutopkgtest workflow."""

    source_artifact_id: int | None = None
    binary_artifacts_ids: list[int] | None = None
    context_artifacts_ids: list[int] | None = None

    qa_suite_id: int | None = None


class LintianWorkflowData(RegressionTrackingWorkflowData):
    """`lintian` workflow data."""

    source_artifact: LookupSingle
    binary_artifacts: LookupMultiple

    vendor: str
    codename: str
    backend: BackendType = BackendType.UNSHARE

    architectures: list[str] | None = None
    arch_all_build_architecture: str = "amd64"
    output: LintianOutput = pydantic.Field(default_factory=LintianOutput)

    include_tags: list[str] = pydantic.Field(default_factory=list)
    exclude_tags: list[str] = pydantic.Field(default_factory=list)
    fail_on_severity: LintianFailOnSeverity = LintianFailOnSeverity.ERROR


class LintianWorkflowDynamicData(BaseDynamicTaskData):
    """Dynamic data for the Lintian workflow."""

    source_artifact_id: int | None = None
    binary_artifacts_ids: list[int] | None = None


class DebDiffWorkflowData(BaseWorkflowData):
    """`debdiff` workflow data."""

    source_artifact: LookupSingle
    binary_artifacts: LookupMultiple

    original: LookupSingle
    extra_flags: list[DebDiffFlags] = pydantic.Field(default_factory=list)

    vendor: str
    codename: str

    arch_all_build_architecture: str = "amd64"


class DebDiffWorkflowDynamicData(BaseDynamicTaskData):
    """Dynamic data for the DebDiff workflow."""

    source_artifact_id: int | None = None
    binary_artifacts_ids: list[int] | None = None

    original_id: int | None = None


class BlhcWorkflowData(RegressionTrackingWorkflowData):
    """`blhc` workflow data."""

    source_artifact: LookupSingle | None = None
    binary_artifacts: LookupMultiple | None = None
    package_build_logs: LookupMultiple | None = None
    extra_flags: list[BlhcFlags] = pydantic.Field(default_factory=list)

    vendor: str
    codename: str

    arch_all_build_architecture: str = "amd64"

    @pydantic.model_validator(mode="after")
    def check_source_artifact(self) -> Self:
        """``source_artifact`` is required for regression tracking."""
        if self.enable_regression_tracking and self.source_artifact is None:
            raise ValueError(
                '"source_artifact" is required if '
                '"enable_regression_tracking" is set'
            )
        return self

    @pydantic.model_validator(mode="after")
    def check_package_build_logs(self) -> Self:
        """We need either ``binary_artifacts`` or ``package_build_logs``."""
        if (self.binary_artifacts is None) == (self.package_build_logs is None):
            raise ValueError(
                "Exactly one of binary_artifacts"
                " or package_build_logs must be set"
            )
        return self


class BlhcWorkflowDynamicData(BaseDynamicTaskData):
    """Dynamic data for Blhc workflow."""

    source_artifact_id: int | None = None
    binary_artifacts_ids: list[int] | None = None
    package_build_logs_ids: list[int] | None = None


class QAFailOn(StrEnum):
    """Conditions to trigger a failure of the whole QA workflow."""

    FAILURE = "failure"
    REGRESSION = "regression"
    NEVER = "never"


class QAWorkflowData(RegressionTrackingWorkflowData):
    """`qa` workflow data."""

    source_artifact: LookupSingle
    binary_artifacts: LookupMultiple
    # Obsolete, but may be present in old rows.
    package_build_logs: LookupMultiple | None = None

    # debian_pipeline sets this if enable_regression_tracking is set, but it
    # wasn't present in older workflows.
    reference_source_artifact: LookupSingle | None = None

    vendor: str
    codename: str
    architectures: list[str] | None = None
    extra_repositories: list[ExtraRepository] | None = None

    architectures_allowlist: list[str] | None = None
    architectures_denylist: list[str] | None = None

    arch_all_build_architecture: str = "amd64"

    enable_autopkgtest: bool = True
    autopkgtest_backend: BackendType = BackendType.AUTO

    enable_reverse_dependencies_autopkgtest: bool = False

    enable_lintian: bool = True
    lintian_backend: BackendType = BackendType.AUTO
    lintian_fail_on_severity: LintianFailOnSeverity = (
        LintianFailOnSeverity.ERROR
    )

    enable_piuparts: bool = True
    piuparts_backend: BackendType = BackendType.AUTO
    piuparts_environment: LookupSingle | None = None

    enable_debdiff: bool = False
    enable_blhc: bool = False

    @staticmethod
    def _fail_on_default_factory(data: dict[str, Any]) -> QAFailOn:
        return (
            QAFailOn.REGRESSION
            if data.get("enable_regression_tracking")
            else QAFailOn.FAILURE
        )

    fail_on: QAFailOn = pydantic.Field(default_factory=_fail_on_default_factory)

    @pydantic.model_validator(mode="after")
    def check_reverse_dependencies_autopkgtest_consistency(self) -> Self:
        """Check consistency of reverse-dependencies-autopkgtest options."""
        if (
            self.enable_reverse_dependencies_autopkgtest
            and self.qa_suite is None
        ):
            raise ValueError(
                '"qa_suite" is required if '
                '"enable_reverse_dependencies_autopkgtest" is set'
            )
        return self

    @pydantic.model_validator(mode="after")
    def enable_debdiff_consistency(self) -> Self:
        """Check consistency of enable_debdiff options."""
        if self.enable_debdiff and self.qa_suite is None:
            raise ValueError(
                '"qa_suite" is required if "enable_debdiff" is set'
            )
        return self


class QAWorkflowDynamicData(BaseDynamicTaskData):
    """Dynamic data for the QA workflow."""

    source_artifact_id: int | None = None
    binary_artifacts_ids: list[int] | None = None
    reference_source_artifact_id: int | None = None
    # Obsolete, but may be present in old rows.
    package_build_logs_ids: list[int] | None = None


class QAFailurePolicy(StrEnum):
    """Policy to apply if the QA workflow fails."""

    IGNORE = "ignore"
    FAIL = "fail"
    CONFIRM = "confirm"


class DebianPipelineWorkflowData(BaseWorkflowData):
    """`debian_pipeline` workflow data."""

    source_artifact: LookupSingle

    suite: LookupSingle | None = None

    vendor: str
    codename: str

    architectures: list[str] | None = None
    architectures_allowlist: list[str] | None = None
    architectures_denylist: list[str] | None = None
    arch_all_build_architecture: str = "amd64"
    extra_repositories: list[ExtraRepository] | None = None

    signing_template_names: dict[str, list[str]] = {}

    sbuild_backend: BackendType = BackendType.AUTO
    sbuild_environment_variant: str | None = None

    qa_suite: LookupSingle | None = None
    regression_tracking_qa_results: LookupSingle | None = None
    enable_regression_tracking: bool = False
    qa_failure_policy: QAFailurePolicy = QAFailurePolicy.IGNORE

    enable_autopkgtest: bool = True
    autopkgtest_backend: BackendType = BackendType.AUTO

    enable_reverse_dependencies_autopkgtest: bool = False

    enable_lintian: bool = True
    lintian_backend: BackendType = BackendType.AUTO
    lintian_fail_on_severity: LintianFailOnSeverity = (
        LintianFailOnSeverity.ERROR
    )

    enable_piuparts: bool = True
    piuparts_backend: BackendType = BackendType.AUTO
    piuparts_environment: LookupSingle | None = None

    enable_debdiff: bool = False

    enable_blhc: bool = True

    enable_confirmation: bool = False

    enable_make_signed_source: bool = False
    make_signed_source_purpose: KeyPurpose | None = None
    make_signed_source_key: str | None = None

    enable_upload: bool = False

    upload_key: str | None = None
    upload_binary_key: str | None = None
    upload_require_signature: bool = True
    upload_include_source: bool = True
    upload_include_binaries: bool = True
    upload_merge_uploads: bool = True
    upload_since_version: str | None = None
    upload_target_distribution: str | None = None
    upload_target: str = (
        "ftp://anonymous@ftp.upload.debian.org/pub/UploadQueue/"
    )
    upload_delayed_days: int | None = None

    publish_replace: bool = False

    @pydantic.model_validator(mode="after")
    def check_reverse_dependencies_autopkgtest_consistency(self) -> Self:
        """Check consistency of reverse-dependencies-autopkgtest options."""
        if (
            self.enable_reverse_dependencies_autopkgtest
            and self.qa_suite is None
        ):
            raise ValueError(
                '"qa_suite" is required if '
                '"enable_reverse_dependencies_autopkgtest" is set'
            )
        return self

    @pydantic.model_validator(mode="after")
    def enable_debdiff_consistency(self) -> Self:
        """Check consistency of enable_debdiff options."""
        if self.enable_debdiff and self.qa_suite is None:
            raise ValueError(
                '"qa_suite" is required if "enable_debdiff" is set'
            )
        return self

    @pydantic.model_validator(mode="after")
    def check_regression_tracking_consistency(self) -> Self:
        """Check consistency of regression-tracking options."""
        if self.enable_regression_tracking:
            if self.qa_suite is None:
                raise ValueError(
                    '"qa_suite" is required if '
                    '"enable_regression_tracking" is set'
                )
            if self.regression_tracking_qa_results is None:
                raise ValueError(
                    '"regression_tracking_qa_results" is required if '
                    '"enable_regression_tracking" is set'
                )
        return self


class DebianPipelineWorkflowDynamicData(BaseDynamicTaskData):
    """Dynamic data for the DebianPipeline workflow."""

    input_source_artifact_id: int | None = None


class PackagePublishWorkflowData(BaseWorkflowData):
    """`package_publish` workflow data."""

    source_artifact: LookupSingle | None = None
    binary_artifacts: LookupMultiple = pydantic.Field(
        default_factory=partial(LookupMultiple, ())
    )
    target_suite: LookupSingle
    unembargo: bool = False
    replace: bool = False
    suite_variables: dict[str, Any] = {}
    update_indexes: bool = True

    @pydantic.model_validator(mode="after")
    def check_one_of_source_or_binary(self) -> Self:
        """Ensure a source or binary artifact is present."""
        if self.source_artifact is None and not self.binary_artifacts:
            raise ValueError(
                '"source_artifact" or "binary_artifacts" must be set'
            )
        return self


class PackagePublishWorkflowDynamicData(BaseDynamicTaskData):
    """Dynamic data for PackagePublish workflow."""

    source_artifact_id: int | None = None
    binary_artifacts_ids: list[int] | None = None
    target_suite_id: int | None = None


class ChildWorkspaceData(BaseWorkflowData):
    """``create_child_workspace`` workflow data."""

    prefix: str | None = pydantic.Field(
        default=None, pattern=workspace_name_regex
    )
    suffix: str = pydantic.Field(pattern=workspace_name_regex)
    public: bool = True
    owner_group: str | None = None
    workflow_template_names: list[str] = []
    expiration_delay: int | None = 60


class UpdateSuiteData(BaseWorkflowData):
    """``update_suite`` workflow data."""

    suite_collection: LookupSingle


class UpdateSuitesData(BaseWorkflowData):
    """``update_suites`` workflow data."""

    force_basic_indexes: bool = False
    only_suites: list[str] | None = None



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common test-helper code involving workflows."""

from collections.abc import Generator
from contextlib import contextmanager
from typing import Any, override
from unittest import mock

from django.db import connection
from django.test import override_settings

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts.models import TaskTypes
from debusine.db.models import Artifact, WorkRequest, WorkflowTemplate
from debusine.server.scheduler import schedule
from debusine.server.tasks.tests.helpers import DBTaskTestCase, SampleBaseDBTask
from debusine.server.workflows import SbuildWorkflow, Workflow, workflow_utils
from debusine.server.workflows.models import BaseWorkflowData
from debusine.tasks.models import BaseDynamicTaskData
from debusine.test.django import TestCase


class SampleWorkflow[WD: BaseWorkflowData, DTD: BaseDynamicTaskData](
    Workflow[WD, DTD], SampleBaseDBTask[WD, DTD]
):
    """Common test implementation of Workflow methods."""


class WorkflowTestCase[TaskClass: Workflow[Any, Any]](
    DBTaskTestCase[TaskClass], TestCase
):
    """Base class for workflow tests."""

    #: Parent WorkRequest for the workflow being tested
    parent: WorkRequest | None = None
    #: Validate data of the workflow being tested
    validate: bool = True
    #: Internal collection artifacts to pass to create_workflow
    internal_collection_artifacts: (
        list[tuple[Artifact, dict[str, Any]]] | None
    ) = None

    @override
    def setUp(self) -> None:
        super().setUp()
        self.template: WorkflowTemplate | None = None

    def create_template(self, name: str, **kwargs: Any) -> WorkflowTemplate:
        """
        Create a workflow template.

        :param kwargs: if any keyword arguments are provided, they replace the
          corresponding values in ``self.task_data``.
        """
        task_data = self.task_data.copy()
        task_data.update(kwargs)
        return self.playground.create_workflow_template(
            name, self.task_class.name, static_parameters=task_data
        )

    @override
    def work_request_create(self, **kwargs: Any) -> WorkRequest:
        task_data = self.task_data.copy()
        task_data.update(kwargs)
        work_request = self.playground.create_workflow(
            task_name=(
                self.template
                if self.template is not None
                else self.task_class.name
            ),
            task_data=self.task_class.task_data_type(**task_data),
            parent=self.parent,
            validate=self.validate,
            internal_collection_artifacts=self.internal_collection_artifacts,
        )
        return work_request

    @classmethod
    def get_workflow(cls, work_request: WorkRequest) -> TaskClass:
        """Create and return a suitable Workflow for this work request."""
        task = work_request.get_task()
        assert isinstance(task, cls.task_class)
        return task

    @contextmanager
    def schedule_and_run(self) -> Generator[list[WorkRequest]]:
        """
        Run the scheduler and any resulting Celery callbacks.

        Yields the list of scheduled work requests so that the caller can
        assert things about them before they are run.
        """
        start_count = len(connection.run_on_commit)
        with (
            mock.patch("pgtransaction.Atomic.execute_set_transaction_modes"),
            override_settings(PGTRANSACTION_RETRY=0),
        ):
            with self.captureOnCommitCallbacks() as on_commit:
                yield schedule()
            with override_settings(CELERY_TASK_ALWAYS_EAGER=True):
                for callback in on_commit:
                    callback()
                del connection.run_on_commit[
                    start_count : start_count + len(on_commit)
                ]

    def schedule_and_run_workflow_callback(self, parent: WorkRequest) -> None:
        """
        Run the scheduler and a workflow callback.

        :param parent: Expect the workflow callback to be a child of this
          workflow.
        """
        with self.schedule_and_run() as [work_request]:
            self.assertEqual(work_request.task_type, TaskTypes.INTERNAL)
            self.assertEqual(work_request.task_name, "workflow")
            self.assertEqual(work_request.parent, parent)

    def schedule_and_run_workflow(self, workflow: WorkRequest) -> None:
        """
        Run the scheduler and a workflow.

        :param workflow: Expect this workflow to be scheduled.
        """
        with self.schedule_and_run() as work_requests:
            self.assertIn(workflow, work_requests)

    def simulate_sbuild_workflow_completion(
        self,
        sbuild_workflow: WorkRequest,
        only_architecture: str | None = None,
        binary_names: list[str] | None = None,
        section: str = "misc",
        priority: str = "optional",
        use_data: bool = True,
    ) -> tuple[dict[str, list[Artifact]], dict[str, Artifact]]:
        """
        Simulate completion of an sbuild workflow.

        :param sbuild_workflow: The workflow to complete.
        :param only_architecture: If given, only complete the sbuild task
          for this architecture.
        :param binary_names: List of binary package names to create;
          defaults to a single binary named after the source package.
        :param section: ``Section`` field for binary artifacts.
        :param priority: ``Priority`` field for binary artifacts.
        :return: A dictionary mapping architectures to the
          ``debian:binary-package`` artifacts created for them, and a
          dictionary mapping architectures to their
          ``debian:package-build-log`` artifacts.
        """
        sbuild_workflow_task = sbuild_workflow.get_task()
        assert isinstance(sbuild_workflow_task, SbuildWorkflow)
        source_data = workflow_utils.source_package_data(
            sbuild_workflow_task, configuration_key="input.source_artifact"
        )

        binaries: dict[str, list[Artifact]] = {}
        build_logs: dict[str, Artifact] = {}
        for sbuild in sbuild_workflow.children.filter(
            task_type=TaskTypes.WORKER,
            task_name="sbuild",
            workflow_data_json__step__startswith="build-",
        ).order_by("workflow_data_json__step"):
            assert sbuild.workflow_data.step is not None
            arch = sbuild.workflow_data.step.removeprefix("build-")
            if only_architecture is not None and arch != only_architecture:
                continue
            upload = self.playground.create_upload_artifacts(
                src_name=source_data.name,
                version=source_data.version,
                binary=True,
                source=False,
                binaries=[
                    (binary_name, arch)
                    for binary_name in (binary_names or (source_data.name,))
                ],
                section=section,
                priority=priority,
                workspace=sbuild.workspace,
                work_request=sbuild,
            )
            binaries[arch] = upload.binaries
            build_logs[arch] = self.playground.create_build_log_artifact(
                source=source_data.name,
                version=source_data.version,
                build_arch=arch,
                workspace=sbuild.workspace,
                work_request=sbuild,
            )
            self.playground.create_artifact_relation(
                artifact=build_logs[arch], target=upload.upload
            )
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS, use_data=use_data
            )
        return binaries, build_logs

    @override
    def build_expected_tag_candidates(
        self, **kwargs: set[str]
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        provided, required = super().build_expected_tag_candidates(**kwargs)
        required.add(
            wtags.ProvenanceRequired.SYSTEM, {wtags.WORKER_TYPE_NOT_ASSIGNABLE}
        )
        return provided, required



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for Debusine Celery configuration."""

import json
from multiprocessing import Lock
from pathlib import Path
from typing import Any, override
from unittest import mock

from celery import shared_task, signals, states
from celery.contrib.testing.worker import start_worker
from celery.utils.nodenames import gethostname, nodename
from django.conf import settings
from django.db import transaction
from django.test import override_settings
from django_celery_results.models import TaskResult

import debusine.worker.tags as wtags
from debusine.db.models import Token, Worker
from debusine.project.celery import (
    CeleryToken,
    app,
    apply_async_on_commit,
    get_celery_token,
    get_celery_worker_metadata,
    get_celery_worker_provided_tags,
    get_scheduler_worker_metadata,
    make_app,
)
from debusine.tasks.models import WorkerType
from debusine.test.django import TestCase, TransactionTestCase

_lock = Lock()


# mypy complains that celery.shared_task is untyped, which is true, but we
# can't fix that here.
@shared_task  # type: ignore[untyped-decorator]
def _locked_noop() -> None:
    """Sample Celery task to do nothing, once a lock is available."""
    with _lock:
        pass


# mypy complains that celery.shared_task is untyped, which is true, but we
# can't fix that here.
@shared_task  # type: ignore[untyped-decorator]
def _locked_reverse_string(argument: str, **kwargs: Any) -> str:
    """Sample Celery task to reverse a string, once a lock is available."""
    with _lock:
        return "".join(reversed(argument))


@override_settings(CELERY_BROKER_URL="memory://")
class TestApplyAsyncOnCommit(TransactionTestCase):
    """Tests for ``apply_async_on_commit`` and ``store_pending_state``."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.app = make_app()

    def test_result_states(self) -> None:
        """The task's state can be tracked in the database."""
        with (
            mock.patch("debusine.django.django_utils.test_data_override"),
            start_worker(self.app, perform_ping_check=False) as celery_worker,
        ):
            expected_task_name = f"{__name__}._locked_reverse_string"
            expected_task_args = ["hello"]
            expected_task_kwargs = {"foo": "bar", "value": 1}

            def assert_task_ok(task_id: str, expected_status: str) -> None:
                task_result = TaskResult.objects.get(task_id=task_id)
                self.assertEqual(task_result.status, expected_status)
                self.assertEqual(task_result.task_name, expected_task_name)
                self.assertEqual(
                    json.loads(task_result.task_args), expected_task_args
                )
                self.assertEqual(
                    json.loads(task_result.task_kwargs), expected_task_kwargs
                )

            # Schedule a task, but with a lock held so that we get a chance
            # to look at its result in the database before it finishes.
            with _lock:
                with transaction.atomic():
                    task_id = apply_async_on_commit(
                        _locked_reverse_string,
                        args=("hello",),
                        kwargs={"foo": "bar", "value": 1},
                    )
                    assert task_id is not None

                    # The result before committing the transaction is correct.
                    assert_task_ok(task_id, states.PENDING)

                # The result after committing the transaction is correct.
                assert_task_ok(task_id, states.PENDING)

            # Release the lock to allow the task to complete.  Its result in
            # the database is updated with the new status.
            self.assertEqual(
                _locked_reverse_string.AsyncResult(task_id).get(), "olleh"
            )

            assert_task_ok(task_id, states.SUCCESS)

            signals.worker_shutdown.send(sender=celery_worker)

    def test_result_states_no_arguments(self) -> None:
        """A task with no arguments is tracked correctly in the database."""
        with (
            mock.patch("debusine.django.django_utils.test_data_override"),
            start_worker(self.app, perform_ping_check=False) as celery_worker,
        ):
            expected_task_name = f"{__name__}._locked_noop"

            def assert_task_ok(task_id: str, expected_status: str) -> None:
                task_result = TaskResult.objects.get(task_id=task_id)
                self.assertEqual(task_result.status, expected_status)
                self.assertEqual(task_result.task_name, expected_task_name)
                self.assertEqual(json.loads(task_result.task_args), [])
                self.assertEqual(json.loads(task_result.task_kwargs), {})

            # Schedule a task, but with a lock held so that we get a chance
            # to look at its result in the database before it finishes.
            with _lock:
                with transaction.atomic():
                    task_id = apply_async_on_commit(_locked_noop)
                    assert task_id is not None

                    # The result before committing the transaction is correct.
                    assert_task_ok(task_id, states.PENDING)

                # The result after committing the transaction is correct.
                assert_task_ok(task_id, states.PENDING)

            # Release the lock to allow the task to complete.  Its result in
            # the database is updated with the new status.
            self.assertIsNone(_locked_noop.AsyncResult(task_id).get())

            assert_task_ok(task_id, states.SUCCESS)

            signals.worker_shutdown.send(sender=celery_worker)

    def test_result_states_once_per_savepoint(self) -> None:
        """``once_per_savepoint=True`` allows batching up tasks."""
        with (
            mock.patch("debusine.django.django_utils.test_data_override"),
            start_worker(self.app, perform_ping_check=False) as celery_worker,
        ):
            expected_task_name = f"{__name__}._locked_noop"

            def assert_task_ok(task_id: str, expected_status: str) -> None:
                task_result = TaskResult.objects.get(task_id=task_id)
                self.assertEqual(task_result.status, expected_status)
                self.assertEqual(task_result.task_name, expected_task_name)
                self.assertEqual(json.loads(task_result.task_args), [])
                self.assertEqual(json.loads(task_result.task_kwargs), {})

            with _lock:
                with transaction.atomic():
                    with transaction.atomic():
                        task_id = apply_async_on_commit(
                            _locked_noop, once_per_savepoint=True
                        )
                        assert task_id is not None
                        assert_task_ok(task_id, states.PENDING)

                        # If the task has already been scheduled in this
                        # savepoint, no new task is scheduled.
                        self.assertIsNone(
                            apply_async_on_commit(
                                _locked_noop, once_per_savepoint=True
                            )
                        )

                    # If the task has not been scheduled in this savepoint,
                    # a new task is scheduled.
                    task_id_2 = apply_async_on_commit(
                        _locked_noop, once_per_savepoint=True
                    )
                    assert task_id_2 is not None
                    self.assertNotEqual(task_id_2, task_id)
                    assert_task_ok(task_id_2, states.PENDING)

                assert_task_ok(task_id, states.PENDING)
                assert_task_ok(task_id_2, states.PENDING)

            self.assertIsNone(_locked_noop.AsyncResult(task_id).get())
            self.assertIsNone(_locked_noop.AsyncResult(task_id_2).get())

            assert_task_ok(task_id, states.SUCCESS)
            assert_task_ok(task_id_2, states.SUCCESS)

            signals.worker_shutdown.send(sender=celery_worker)


class TestCeleryToken(TestCase):
    """Tests for :py:class:`CeleryToken`."""

    def test_existing_token(self) -> None:
        tokens_dir = self.create_temporary_directory()
        key = "existing-key"
        Path(tokens_dir, "test").write_text(key)

        with override_settings(
            DEBUSINE_FQDN="debusine.example.org",
            DEBUSINE_TOKENS_DIRECTORY=str(tokens_dir),
        ):
            celery_token = CeleryToken("test")
        self.addCleanup(celery_token.token.delete)

        self.assertEqual(celery_token.token.hash, Token._generate_hash(key))
        self.assertEqual(
            celery_token.client.base_api_url, "https://debusine.example.org/api"
        )
        self.assertEqual(celery_token.client.token, key)
        self.assertIsNone(celery_token.client.scope)

    def test_new_token(self) -> None:
        tokens_dir = self.create_temporary_directory()

        with override_settings(
            DEBUSINE_FQDN="debusine.example.org",
            DEBUSINE_TOKENS_DIRECTORY=str(tokens_dir),
        ):
            celery_token = CeleryToken("test")
        self.addCleanup(celery_token.token.delete)

        key = Path(tokens_dir, "test").read_text()
        self.assertEqual(len(key), 64)
        self.assertEqual(celery_token.token.hash, Token._generate_hash(key))
        self.assertEqual(
            celery_token.client.base_api_url, "https://debusine.example.org/api"
        )
        self.assertEqual(celery_token.client.token, key)
        self.assertIsNone(celery_token.client.scope)


class TestWorkerMetadata(TestCase):
    """Test for worker metadata generation functions."""

    def test_celery_tags(self) -> None:
        system_provided_tags = self.enterContext(
            self.spy("debusine.worker.system_information.system_provided_tags")
        )
        get_provided_celery_worker_tags = self.enterContext(
            mock.patch(
                "debusine.server.tasks.get_provided_celery_worker_tags",
                return_value={"servertasks"},
            )
        )
        executor_provided_tags = self.enterContext(
            mock.patch(
                "debusine.tasks.executors.executor_provided_tags",
                return_value={"executor"},
            )
        )
        tags = get_celery_worker_provided_tags()
        self.assertIn(wtags.WORKER_TYPE_CELERY, tags)
        self.assertNotIn("executor", tags)
        self.assertIn("servertasks", tags)

        system_provided_tags.assert_called_once_with(WorkerType.CELERY)
        executor_provided_tags.assert_not_called()
        get_provided_celery_worker_tags.assert_called_once_with()

    def test_celery(self) -> None:
        system_metadata = self.enterContext(
            self.spy("debusine.worker.system_information.system_metadata")
        )
        analyze_celery_worker_tasks = self.enterContext(
            self.spy("debusine.server.tasks.analyze_celery_worker_tasks")
        )
        analyze_worker_all_executors = self.enterContext(
            self.spy("debusine.tasks.executors.analyze_worker_all_executors")
        )
        get_celery_worker_provided_tags = self.enterContext(
            self.spy("debusine.project.celery.get_celery_worker_provided_tags")
        )

        md = get_celery_worker_metadata()
        self.assertEqual(md["system:worker_type"], WorkerType.CELERY)
        self.assertIn("provided_tags", md)

        system_metadata.assert_called_once_with(WorkerType.CELERY)
        analyze_celery_worker_tasks.assert_called_once_with()
        analyze_worker_all_executors.assert_not_called()
        get_celery_worker_provided_tags.assert_called_once_with()

    def test_scheduler(self) -> None:
        system_metadata = self.enterContext(
            self.spy("debusine.worker.system_information.system_metadata")
        )
        md = get_scheduler_worker_metadata()
        self.assertEqual(md["system:worker_type"], WorkerType.SCHEDULER)
        self.assertNotIn("provided_tags", md)

        system_metadata.assert_called_once_with(WorkerType.SCHEDULER)


@override_settings(CELERY_BROKER_URL="memory://")
class TestConnectWorker(TransactionTestCase):
    """Tests for :py:func:`connect_worker`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.app = make_app()
        Worker.objects.filter(
            worker_type__in={WorkerType.CELERY, WorkerType.SCHEDULER}
        ).delete()

    def test_server_task_worker(self) -> None:
        """The main worker creates a Worker instance that runs server tasks."""
        with (
            mock.patch("debusine.django.django_utils.test_data_override"),
            # Set a concurrency that's unlikely to be the actual number of
            # CPUs we have.
            start_worker(
                self.app,
                concurrency=5,
                perform_ping_check=False,
                hostname=nodename("celery", gethostname()),
            ) as celery_worker,
        ):
            worker = Worker.objects.get(worker_type=WorkerType.CELERY)
            self.assertEqual(worker.concurrency, 5)
            self.assertEqual(
                worker.dynamic_metadata["system:worker_type"], WorkerType.CELERY
            )
            self.assertIn("provided_tags", worker.dynamic_metadata)
            self.assertIsNotNone(worker.connected_at)
            self.assertFalse(
                Worker.objects.filter(worker_type=WorkerType.SCHEDULER).exists()
            )

            token = Token.objects.get_token_or_none(
                Path(settings.DEBUSINE_TOKENS_DIRECTORY, "celery").read_text()
            )
            assert token is not None
            self.assertEqual(token.worker, worker)
            self.assertTrue(token.enabled)
            celery_token = get_celery_token()
            assert celery_token is not None
            self.assertEqual(celery_token.token, token)

            signals.worker_shutdown.send(sender=celery_worker)

            worker.refresh_from_db()
            self.assertIsNone(worker.connected_at)
            self.assertFalse(
                Worker.objects.filter(worker_type=WorkerType.SCHEDULER).exists()
            )

    def test_server_task_worker_token_already_exists(self) -> None:
        with transaction.atomic():
            existing_worker = Worker.objects.get_or_create_celery()
            existing_worker.token = CeleryToken("celery").token
            existing_worker.save()

        with (
            mock.patch("debusine.django.django_utils.test_data_override"),
            # Set a concurrency that's unlikely to be the actual number of
            # CPUs we have.
            start_worker(
                self.app,
                concurrency=5,
                perform_ping_check=False,
                hostname=nodename("celery", gethostname()),
            ) as celery_worker,
        ):
            worker = Worker.objects.get(worker_type=WorkerType.CELERY)
            self.assertEqual(worker, existing_worker)

            token = Token.objects.get_token_or_none(
                Path(settings.DEBUSINE_TOKENS_DIRECTORY, "celery").read_text()
            )
            assert token is not None
            self.assertEqual(token.worker, worker)
            self.assertTrue(token.enabled)
            self.assertEqual(token, existing_worker.token)
            celery_token = get_celery_token()
            assert celery_token is not None
            self.assertEqual(celery_token.token, token)

            signals.worker_shutdown.send(sender=celery_worker)

    def test_scheduler_worker(self) -> None:
        """The scheduler worker creates a Worker instance."""
        with (
            mock.patch("debusine.django.django_utils.test_data_override"),
            start_worker(
                app,
                concurrency=1,
                perform_ping_check=False,
                hostname=nodename("scheduler", gethostname()),
            ) as celery_worker,
        ):
            worker = Worker.objects.get(worker_type=WorkerType.SCHEDULER)
            self.assertEqual(worker.concurrency, 1)
            self.assertEqual(
                worker.dynamic_metadata["system:worker_type"],
                WorkerType.SCHEDULER,
            )
            self.assertIsNotNone(worker.connected_at)
            self.assertFalse(
                Worker.objects.filter(worker_type=WorkerType.CELERY).exists()
            )

            token = Token.objects.get_token_or_none(
                Path(
                    settings.DEBUSINE_TOKENS_DIRECTORY, "scheduler"
                ).read_text()
            )
            assert token is not None
            self.assertEqual(token.worker, worker)
            self.assertTrue(token.enabled)
            celery_token = get_celery_token()
            assert celery_token is not None
            self.assertEqual(celery_token.token, token)

            signals.worker_shutdown.send(sender=celery_worker)

            worker.refresh_from_db()
            self.assertIsNone(worker.connected_at)
            self.assertFalse(
                Worker.objects.filter(worker_type=WorkerType.CELERY).exists()
            )

    def test_scheduler_worker_token_already_exists(self) -> None:
        with transaction.atomic():
            existing_worker = Worker.objects.get_or_create_scheduler()
            existing_worker.token = CeleryToken("scheduler").token
            existing_worker.save()

        with (
            mock.patch("debusine.django.django_utils.test_data_override"),
            start_worker(
                app,
                concurrency=1,
                perform_ping_check=False,
                hostname=nodename("scheduler", gethostname()),
            ) as celery_worker,
        ):
            worker = Worker.objects.get(worker_type=WorkerType.SCHEDULER)
            self.assertEqual(worker, existing_worker)

            token = Token.objects.get_token_or_none(
                Path(
                    settings.DEBUSINE_TOKENS_DIRECTORY, "scheduler"
                ).read_text()
            )
            assert token is not None
            self.assertEqual(token.worker, worker)
            self.assertTrue(token.enabled)
            self.assertEqual(token, existing_worker.token)
            celery_token = get_celery_token()
            assert celery_token is not None
            self.assertEqual(celery_token.token, token)

            signals.worker_shutdown.send(sender=celery_worker)

    def test_provisioner_worker(self) -> None:
        """The provisioner worker does not create a Worker instance."""
        with (
            mock.patch("debusine.django.django_utils.test_data_override"),
            start_worker(
                app,
                concurrency=1,
                perform_ping_check=False,
                hostname=nodename("provisioner", gethostname()),
            ) as celery_worker,
        ):
            self.assertFalse(
                Worker.objects.filter(
                    worker_type__in={WorkerType.CELERY, WorkerType.SCHEDULER}
                ).exists()
            )
            self.assertIsNone(get_celery_token())

            signals.worker_shutdown.send(sender=celery_worker)

            self.assertFalse(
                Worker.objects.filter(
                    worker_type__in={WorkerType.CELERY, WorkerType.SCHEDULER}
                ).exists()
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application: workflows."""

import logging
from typing import Any, Literal, override

from django.core.exceptions import PermissionDenied
from rest_framework import mixins, status
from rest_framework.exceptions import PermissionDenied as DRFPermissionDenied
from rest_framework.generics import get_object_or_404
from rest_framework.parsers import JSONParser
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import BaseSerializer

from debusine.client.models import RuntimeParameter
from debusine.db.context import context
from debusine.db.models import Group, WorkRequest, WorkflowTemplate, Workspace
from debusine.db.models.permissions import anonymous_access
from debusine.db.models.work_requests import WorkflowTemplateQuerySet
from debusine.server.exceptions import DebusineAPIException
from debusine.server.serializers import (
    CreateWorkflowRequestSerializer,
    WorkRequestSerializer,
    WorkflowTemplateSerializer,
)
from debusine.server.views.base import (
    BaseAPIView,
    CanDisplayFilterBackend,
    GenericViewSetBase,
)
from debusine.server.views.rest import IsUserTokenAuthenticated

logger = logging.getLogger(__name__)


class WorkflowTemplateViewSet(
    BaseAPIView,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.UpdateModelMixin,
    mixins.DestroyModelMixin,
    GenericViewSetBase[WorkflowTemplate],
):
    """Return, create and delete workflow templates."""

    permission_classes = [IsUserTokenAuthenticated]
    serializer_class = WorkflowTemplateSerializer
    filter_backends = [CanDisplayFilterBackend]
    parser_classes = [JSONParser]

    @override
    def initial(self, request: Request, *args: Any, **kwargs: Any) -> None:
        """Return the initial request object."""
        super().initial(request, *args, **kwargs)
        self.set_current_workspace(kwargs["workspace"])

    @override
    def get_queryset(self) -> WorkflowTemplateQuerySet[Any]:
        """Get the query set for this view."""
        self.enforce(anonymous_access)  # Filtered to can_display
        return (
            WorkflowTemplate.objects.can_display(context.pc)
            .filter(workspace=context.require_workspace())
            .order_by("name")
        )

    @override
    def get_object(self) -> WorkflowTemplate:
        """Return the object the view is displaying."""
        queryset = self.filter_queryset(self.get_queryset())
        pk = self.kwargs["pk"]
        if pk.isdigit():
            obj = get_object_or_404(queryset, pk=int(pk))
        else:
            obj = get_object_or_404(queryset, name=pk)
        assert isinstance(obj, WorkflowTemplate)
        # May raise a permission denied
        self.check_object_permissions(self.request, obj)
        self.enforce(obj.can_display)
        return obj

    def _check_permissions(
        self, serializer: BaseSerializer[WorkflowTemplate]
    ) -> None:
        """Check permissions for various special cases."""
        # Only users who can create workspaces in this scope may
        # create/update workflow templates for create_child_workspace, since
        # that amounts to delegating that authority.
        task_name = serializer.validated_data.get("task_name")
        if task_name == "create_child_workspace":
            self.enforce(context.require_scope().can_create_workspace)

        # Only users with appropriate permissions may set priorities.
        priority = serializer.validated_data.get("priority")
        if priority is not None and priority > 0:
            token = self.request.auth
            # TODO: This should be replaced by self.enforce with some
            # appropriate debusine permission.
            if (
                token is None
                or token.user is None
                or not token.user.has_perm("db.manage_workrequest_priorities")
            ):
                raise DRFPermissionDenied(
                    "You are not permitted to set positive priorities"
                )

        # Ensure extra groups can be granted
        extra_groups = serializer.validated_data.get("extra_groups", ())
        for group in extra_groups:
            if not group.can_grant(context.pc):
                raise DRFPermissionDenied(
                    f"You are not allowed to grant extra group {group}"
                )

    def _ensure_unique_name(
        self, workspace: Workspace, name: str, *, title: str
    ) -> None:
        if WorkflowTemplate.objects.filter(
            workspace=workspace,
            name=name,
        ).exists():
            raise DebusineAPIException(
                title=title,
                detail=f"A workflow template called {name!r} already exists",
                status_code=status.HTTP_400_BAD_REQUEST,
            )

    def set_name(self, instance: WorkflowTemplate, name: str) -> bool:
        """
        Rename the workflow template.

        :return: True if the name has been changed
        """
        if name == instance.name:
            return False

        self._ensure_unique_name(
            instance.workspace,
            name,
            title="New name conflicts with an existing workflow template",
        )
        instance.name = name
        return True

    def set_priority(self, instance: WorkflowTemplate, value: int) -> bool:
        """
        Set the priority.

        :return: True if the value has been changed
        """
        if value == instance.priority:
            return False
        instance.priority = value
        return True

    def set_static_parameters(
        self, instance: WorkflowTemplate, value: dict[str, Any]
    ) -> bool:
        """
        Set the static parameters.

        :return: True if the value has been changed
        """
        if value == instance.static_parameters:
            return False
        # TODO: The key-value data should be validated, once we can (See #818)
        instance.static_parameters = value
        return True

    def set_runtime_parameters(
        self,
        instance: WorkflowTemplate,
        value: dict[str, Any] | Literal[RuntimeParameter.ANY],
    ) -> bool:
        """
        Set the runtime parameters.

        :return: True if the value has been changed
        """
        if value == instance.runtime_parameters:
            return False
        # TODO: Like static_parameters above, we should perform full validation.
        instance.runtime_parameters = value
        return True

    def set_extra_groups(
        self, instance: WorkflowTemplate, groups: list[Group]
    ) -> bool:
        """
        Set extra groups.

        :return: True if the value has been changed
        """
        if sorted(
            instance.extra_groups.all().values_list("pk", flat=True)
        ) == sorted(g.pk for g in groups):
            return False
        instance.extra_groups.set(groups)
        return True

    def set_restricted(self, instance: WorkflowTemplate, value: bool) -> bool:
        """
        Set the restricted flag.

        :return: True if the value has been changed
        """
        if value == instance.restricted:
            return False
        instance.restricted = value
        return True

    @override
    def perform_create(
        self, serializer: BaseSerializer[WorkflowTemplate]
    ) -> None:
        """Create a workflow template."""
        workspace = context.require_workspace()
        self.enforce(workspace.can_configure)
        self._check_permissions(serializer)
        name = serializer.validated_data["name"]
        self._ensure_unique_name(
            workspace,
            name,
            title="Workflow template already exists",
        )
        serializer.save(workspace=workspace)

    @override
    def perform_update(
        self, serializer: BaseSerializer[WorkflowTemplate]
    ) -> None:
        """Update a workflow template."""
        assert isinstance(serializer, WorkflowTemplateSerializer)
        instance = serializer.instance
        assert isinstance(instance, WorkflowTemplate)
        self.enforce(instance.can_configure)
        self._check_permissions(serializer)

        # TODO: currently we are enforcing WorkflowTemplate.can_configure,
        # which is true for workspace owners. This would potentially allow a
        # workspace owner to change a workflow template that grants extra
        # groups that they cannot access. See
        # https://salsa.debian.org/freexian-team/debusine/-/work_items/1480
        # for a discussion

        changed: bool = False
        if "name" in serializer.validated_data:
            changed = (
                self.set_name(instance, serializer.validated_data["name"])
                or changed
            )
        if "task_name" in serializer.validated_data:
            if serializer.validated_data["task_name"] != instance.task_name:
                raise DebusineAPIException(
                    title="The task name of a workflow template"
                    " cannot be changed",
                    status_code=status.HTTP_400_BAD_REQUEST,
                )
        if "priority" in serializer.validated_data:
            changed = (
                self.set_priority(
                    instance,
                    serializer.validated_data["priority"],
                )
                or changed
            )
        if "static_parameters" in serializer.validated_data:
            changed = (
                self.set_static_parameters(
                    instance, serializer.validated_data["static_parameters"]
                )
                or changed
            )
        if "runtime_parameters" in serializer.validated_data:
            changed = (
                self.set_runtime_parameters(
                    instance, serializer.validated_data["runtime_parameters"]
                )
                or changed
            )
        if "extra_groups" in serializer.validated_data:
            changed = (
                self.set_extra_groups(
                    instance, serializer.validated_data["extra_groups"]
                )
                or changed
            )
        if "restricted" in serializer.validated_data:
            changed = (
                self.set_restricted(
                    instance, serializer.validated_data["restricted"]
                )
                or changed
            )

        if changed:
            instance.save()


class WorkflowView(BaseAPIView):
    """Create workflows from a template."""

    permission_classes = [IsUserTokenAuthenticated]
    filter_backends = [CanDisplayFilterBackend]

    def post(self, request: Request) -> Response:
        """Create a new workflow from a template."""
        serializer = CreateWorkflowRequestSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        workspace = serializer.validated_data["workspace"]
        self.set_current_workspace(workspace)
        # TODO: Replace with can_create_workflows, when it exists
        self.enforce(workspace.can_create_work_requests)
        # WorkRequest.objects.create_workflow enforces some more detailed
        # permissions depending on its arguments.

        template_name = serializer.validated_data["template_name"]
        task_name = serializer.validated_data["task_name"]

        if template_name:
            try:
                template = WorkflowTemplate.objects.get(
                    name=template_name, workspace=workspace
                )
            except WorkflowTemplate.DoesNotExist:
                raise DebusineAPIException(
                    title="Workflow template not found",
                    status_code=status.HTTP_404_NOT_FOUND,
                )
            assert not task_name
            task_name = None
        else:
            template = None
            assert task_name

        try:
            workflow = WorkRequest.objects.create_workflow(
                template=template,
                workspace=workspace,
                task_name=task_name,
                data=serializer.validated_data.get("task_data", {}),
            )
            workflow.mark_pending()
        except PermissionDenied as e:
            raise DebusineAPIException(
                title=str(e), status_code=status.HTTP_403_FORBIDDEN
            )
        except Exception as e:
            raise DebusineAPIException(
                title="Cannot create workflow", detail=str(e)
            )

        return Response(
            WorkRequestSerializer(workflow).data, status=status.HTTP_201_CREATED
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""The collection manager for debian:archive collections."""

import datetime as dt
import re
from typing import Any, override

from debian.debian_support import Version
from django.db import IntegrityError, connection
from django.db.models.fields.json import KT
from django.utils import timezone

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    CollectionItemMatchConstraint,
    User,
    WorkRequest,
)
from debusine.server.collections.base import (
    CollectionManagerInterface,
    ItemAdditionError,
)


class DebianArchiveManager(CollectionManagerInterface):
    """Manage collection of category debian:archive."""

    COLLECTION_CATEGORY = CollectionCategory.ARCHIVE
    VALID_ARTIFACT_CATEGORIES = frozenset({ArtifactCategory.REPOSITORY_INDEX})
    VALID_COLLECTION_CATEGORIES = frozenset({CollectionCategory.SUITE})

    @override
    def do_add_artifact(
        self,
        artifact: Artifact,
        *,
        user: User,
        workflow: WorkRequest | None = None,
        variables: dict[str, Any] | None = None,
        name: str | None = None,
        replace: bool = False,
        created_at: dt.datetime | None = None,
        replaced_by: CollectionItem | None = None,
    ) -> CollectionItem:
        if "path" not in (variables or {}):
            raise ItemAdditionError(
                f'Adding {artifact.category} to {CollectionCategory.ARCHIVE} '
                f'requires "path"'
            )
        assert variables is not None

        path = variables["path"]
        if path.startswith("dists/"):
            raise ItemAdditionError(
                "Archive-level repository indexes may not have paths under "
                "dists/"
            )
        data = {"path": path}
        name = "index:{path}".format(**data)

        if replace:
            self.remove_items_by_name(
                name=name,
                child_types=[CollectionItem.Types.ARTIFACT],
                user=user,
                workflow=workflow,
            )

        try:
            return CollectionItem.objects.create_from_artifact(
                artifact,
                parent_collection=self.collection,
                name=name,
                data=data,
                created_at=created_at,
                created_by_user=user,
                created_by_workflow=workflow,
                replaced_by=replaced_by,
            )
        except IntegrityError as exc:
            raise ItemAdditionError(str(exc))

    @override
    def do_add_collection(
        self,
        collection: Collection,
        *,
        user: User,
        workflow: WorkRequest | None = None,
        variables: dict[str, Any] | None = None,  # noqa: ARG002, U100
        name: str | None = None,  # noqa: ARG002, U100
        replace: bool = False,  # noqa: ARG002, U100
        created_at: dt.datetime | None = None,
        replaced_by: CollectionItem | None = None,
    ) -> CollectionItem:
        # Require suites to be in the same workspace as the archive.  Since
        # archives are singleton collections, this has the effect of
        # ensuring that a suite can be in at most one archive.
        if collection.workspace != self.collection.workspace:
            assert collection.category == CollectionCategory.SUITE
            raise ItemAdditionError(
                "Suite must be in the same workspace as the archive"
            )

        try:
            item = CollectionItem.objects.create_from_collection(
                collection,
                parent_collection=self.collection,
                name=collection.name,
                data={},
                created_at=created_at,
                created_by_user=user,
                created_by_workflow=workflow,
                replaced_by=replaced_by,
            )
            # The pool/ directory is shared between suites in the same
            # archive, so we need to copy the suite-level constraints
            # created by DebianSuiteManager.make_constraints to ensure that
            # different suites in the same archive don't contain the same
            # pool file with different contents.
            #
            # bulk_create doesn't support subqueries, so this needs raw SQL
            # in order to perform reasonably if the new suite already
            # contains many packages.
            if item.removed_at is None:
                with connection.cursor() as cursor:
                    cursor.execute(
                        """\
                        INSERT INTO db_collectionitemmatchconstraint
                            (collection_id, collection_item_id, constraint_name,
                             key, value)
                        SELECT
                            %s, collection_item_id, constraint_name, key, value
                        FROM db_collectionitemmatchconstraint
                        WHERE
                            collection_id = %s
                            AND constraint_name = 'pool-file'
                        """,
                        [self.collection.id, collection.id],
                    )
            return item
        except IntegrityError as exc:
            raise ItemAdditionError(str(exc))

    @override
    def do_remove_item(
        self,
        item: CollectionItem,
        *,
        user: User | None = None,
        workflow: WorkRequest | None = None,
    ) -> None:
        item.removed_by_user = user
        item.removed_by_workflow = workflow
        item.removed_at = timezone.now()
        item.save()
        if item.collection is not None and self.collection.data.get(
            "may_reuse_versions", False
        ):
            CollectionItemMatchConstraint.objects.filter(
                collection_item_id__in=item.collection.child_items.all(),
                constraint_name="pool-file",
            ).delete()

    @override
    def do_lookup(self, query: str) -> CollectionItem | None:
        """
        Return one CollectionItem based on the query.

        :param query: `source:NAME`, `source-version:NAME_VERSION`,
          `binary:NAME_ARCHITECTURE`,
          `binary-version:NAME_VERSION_ARCHITECTURE`, or `index:PATH`.  If
          more than one possible CollectionItem matches the query (which is
          possible for `source:` and `binary:` queries): return the one with
          the highest version.
        """
        qs = CollectionItem.objects.active().filter(
            child_type=CollectionItem.Types.ARTIFACT
        )
        via_suite: bool | str = True
        sort_by_version = True

        if m := re.match(r"^source:(.+)$", query):
            qs = qs.annotate(package=KT("data__package")).filter(
                category=ArtifactCategory.SOURCE_PACKAGE,
                package=m.group(1),
            )
        elif m := re.match(r"^source-version:(.+?)_(.+)$", query):
            qs = qs.annotate(
                package=KT("data__package"), version=KT("data__version")
            ).filter(
                category=ArtifactCategory.SOURCE_PACKAGE,
                package=m.group(1),
                version=m.group(2),
            )
        elif m := re.match(r"^binary:(.+?)_(.+)$", query):
            qs = qs.annotate(
                package=KT("data__package"),
                architecture=KT("data__architecture"),
            ).filter(
                category=ArtifactCategory.BINARY_PACKAGE,
                package=m.group(1),
                architecture__in={m.group(2), "all"},
            )
        elif m := re.match(r"^binary-version:(.+?)_(.+?)_(.+)$", query):
            qs = qs.annotate(
                package=KT("data__package"),
                version=KT("data__version"),
                architecture=KT("data__architecture"),
            ).filter(
                category=ArtifactCategory.BINARY_PACKAGE,
                package=m.group(1),
                version=m.group(2),
                architecture__in={m.group(3), "all"},
            )
        elif m := re.match(r"^index:dists/([^/]+)/(.+)$", query):
            qs = qs.annotate(path=KT("data__path")).filter(
                child_type=CollectionItem.Types.ARTIFACT,
                category=ArtifactCategory.REPOSITORY_INDEX,
                path=m.group(2),
            )
            via_suite = m.group(1)
            sort_by_version = False
        elif m := re.match(r"^index:(.+)$", query):
            qs = qs.annotate(path=KT("data__path")).filter(
                child_type=CollectionItem.Types.ARTIFACT,
                category=ArtifactCategory.REPOSITORY_INDEX,
                path=m.group(1),
            )
            via_suite = False
            sort_by_version = False
        else:
            raise LookupError(f'Unexpected lookup format: "{query}"')

        # Ideally the sorting would be done in the database, but that
        # requires the debversion extension and installing that requires a
        # superuser, which is cumbersome since it means we can't do it in a
        # Django migration as currently configured.  We don't expect a
        # single suite to contain many active versions of a single package,
        # so doing the sorting in Python should be fine in practice.
        if via_suite:
            suite_qs = (
                self.collection.child_items.active()
                .filter(
                    child_type=CollectionItem.Types.COLLECTION,
                    category=CollectionCategory.SUITE,
                )
                .values("collection")
            )
            if isinstance(via_suite, str):
                suite_qs = suite_qs.filter(name=via_suite)
            qs = qs.filter(
                parent_collection__in=suite_qs,
                # Technically redundant, but helps PostgreSQL to use the
                # correct index where available (e.g. db_ci_suite_source_idx).
                parent_category=CollectionCategory.SUITE,
            )
        else:
            qs = qs.filter(
                parent_collection=self.collection,
                # Technically redundant, but helps PostgreSQL to use the
                # correct index where available (e.g.
                # db_ci_archive_index_path_idx).
                parent_category=CollectionCategory.ARCHIVE,
            )
        items = list(qs)
        if sort_by_version:
            items = sorted(
                items, key=lambda item: Version(item.data["version"])
            )
        if not items:
            return None
        return items[-1]
